"""
Commands for the Evscaperoom. This contains all in-room commands as well as
admin debug-commands to help development.

Gameplay commands

- `look` - custom look
- `focus` - focus on object (to perform actions on it)
- `<action> <obj>` - arbitrary interaction with focused object
- `stand` - stand on the floor, resetting any position
- `emote` - free-form emote
- `say/whisper/shout` - simple communication

Other commands

- `evscaperoom` - starts the evscaperoom top-level menu
- `help` - custom in-room help command
- `options` - set game/accessibility options
- `who` - show who's in the room with you
- `quit` - leave a room, return to menu

Admin/development commands

- `jumpstate` - jump to specific room state
- `flag` - assign a flag to an object
- `createobj` - create a room-object set up for Evscaperoom

"""

import re

from django.conf import settings

from evennia import (
    SESSION_HANDLER,
    CmdSet,
    Command,
    InterruptCommand,
    default_cmds,
    syscmdkeys,
)
from evennia.utils import variable_from_module

from .utils import create_evscaperoom_object

_AT_SEARCH_RESULT = variable_from_module(*settings.SEARCH_AT_RESULT.rsplit(".", 1))

_RE_ARGSPLIT = re.compile(r"\s(with|on|to|in|at)\s", re.I + re.U)
_RE_EMOTE_SPEECH = re.compile(r"(\".*?\")|(\'.*?\')")
_RE_EMOTE_NAME = re.compile(r"(/\w+)")
_RE_EMOTE_PROPER_END = re.compile(r"\.$|\.[\'\"]$|\![\'\"]$|\?[\'\"]$")


# configurable help

if hasattr(settings, "EVSCAPEROOM_HELP_SUMMARY_TEXT"):
    _HELP_SUMMARY_TEXT = settings.EVSCAPEROOM_HELP_SUMMARY_TEXT
else:
    _HELP_SUMMARY_TEXT = """
   |yWhat to do ...|n
    - Your goal is to |wescape|n the room. To do that you need to |wlook|n at
      your surroundings for clues on how to escape. When you find something
      interesting, |wexamine|n it for any actions you could take with it.
   |yHow to explore ...|n
    - |whelp [obj or command]|n           - get usage help (never puzzle-related)
    - |woptions|n                         - set game/accessibility options
    - |wlook/l [obj]|n                    - give things a cursory look.
    - |wexamine/ex/e [obj]|n              - look closer at an object. Use again to
                                        look away.
    - |wstand|n                           - stand up if you were sitting, lying etc.
   |yHow to express yourself ...|n
    - |wwho [all]|n                       - show who's in the room or on server.
    - |wemote/pose/: <something>|n        - free-form emote. Use /me to refer
                                        to yourself and /name to refer to other
                                        things/players. Use quotes "..." to speak.
    - |wsay/; <something>|n               - quick-speak your mind
    - |wwhisper <something>|n             - whisper your mind
    - |wshout <something>|n               - shout your mind
   |yHow to quit like a little baby ...|n
    - |wquit / give up|n                  - admit defeat and give up
"""

_HELP_FALLBACK_TEXT = """
There is no help to be had about |y{this}|n. To look away, use |wexamine|n on
its own or with another object you are interested in.
"""

_QUIT_WARNING = """
|rDo you really want to quit?|n

{warning}

Enter |w'quit'|n again to truly give up.
"""

_QUIT_WARNING_CAN_COME_BACK = """
(Since you are not the last person to leave this room, you |gcan|n get back in here
by joining room '|c{roomname}|n' from the menu. Note however that if you leave
now, any personal achievements you may have gathered so far will be |rlost|n!)
"""

_QUIT_WARNING_LAST_CHAR = """
(You are the |rlast|n player to leave this room ('|c{roomname}|n'). This means that when you
leave, this room will go away and you |rwon't|n be able to come back to it!)
"""


class CmdEvscapeRoom(Command):
    """
    Base command parent for all Evscaperoom commands.

    This operates on the premise of 'focus' - the 'focus'
    is set on the caller, then subsequent commands will
    operate on that focus. If no focus is set,
    the operation will be general or on the room.

    Syntax:

       command [<obj1>|<arg1>] [<prep> <obj2>|<arg2>]

    """

    # always separate the command from any args with a space
    arg_regex = r"(/\w+?(\s|$))|\s|$"
    help_category = "Evscaperoom"

    # these flags allow child classes to determine how strict parsing for obj1/obj2 should be
    # (if they are given at all):
    # True - obj1/obj2 must be found as Objects, otherwise it's an error aborting command
    # False - obj1/obj2 will remain None, instead self.arg1, arg2 will be stored as strings
    # None - if obj1/obj2 are found as Objects, set them, otherwise set arg1/arg2 as strings
    obj1_search = None
    obj2_search = None

    def _search(self, query, required):
        """
        This implements the various search modes

        Args:
            query (str): The search query
            required (bool or None): This defines if query *must* be
                found to match a single local Object or not. If None,
                a non-match means returning the query unchanged. When
                False, immediately return the query. If required is False,
                don't search at all.
        Return:
            match (Object or str): The match or the search string depending
                on the `required` mode.
        Raises:
            InterruptCommand: Aborts the command quietly.
        Notes:
            The _AT_SEARCH_RESULT function will handle all error messaging
            for us.

        """
        if required is False:
            return None, query

        matches = self.caller.search(query, quiet=True)

        if not matches or len(matches) > 1:
            if required:
                if not query:
                    self.caller.msg("You must give an argument.")
                else:
                    _AT_SEARCH_RESULT(matches, self.caller, query=query)
                raise InterruptCommand
            else:
                return None, query
        else:
            return matches[0], None

    def parse(self):
        """
        Parse incoming arguments for use in all child classes.

        """
        caller = self.caller
        self.args = self.args.strip()

        # splits to either ['obj'] or e.g. ['obj', 'on', 'obj']
        parts = [part.strip() for part in _RE_ARGSPLIT.split(" " + self.args, 1)]
        nparts = len(parts)
        self.obj1 = None
        self.arg1 = None
        self.prep = None
        self.obj2 = None
        self.arg2 = None
        if nparts == 1:
            self.obj1, self.arg1 = self._search(parts[0], self.obj1_search)
        elif nparts == 3:
            obj1, self.prep, obj2 = parts
            self.obj1, self.arg1 = self._search(obj1, self.obj1_search)
            self.obj2, self.arg2 = self._search(obj2, self.obj2_search)

        self.room = caller.location
        self.roomstate = self.room.db.state

    @property
    def focus(self):
        return self.caller.attributes.get("focus", category=self.room.db.tagcategory)

    @focus.setter
    def focus(self, obj):
        self.caller.attributes.add("focus", obj, category=self.room.tagcategory)

    @focus.deleter
    def focus(self):
        self.caller.attributes.remove("focus", category=self.room.tagcategory)


class CmdGiveUp(CmdEvscapeRoom):
    """
    Give up

    Usage:
      give up

    Abandons your attempts at escaping and of ever winning the pie-eating contest.

    """

    key = "give up"
    aliases = ("abort", "chicken out", "quit", "q")

    def func(self):
        from .menu import run_evscaperoom_menu

        nchars = len(self.room.get_all_characters())
        if nchars == 1:
            warning = _QUIT_WARNING_LAST_CHAR.format(roomname=self.room.name)
            warning = _QUIT_WARNING.format(warning=warning)
        else:
            warning = _QUIT_WARNING_CAN_COME_BACK.format(roomname=self.room.name)
            warning = _QUIT_WARNING.format(warning=warning)

        ret = yield (warning)
        if ret.upper() == "QUIT":
            self.msg("|R ... Oh. Okay then. Off you go.|n\n")
            yield (1)

            self.room.log(f"QUIT: {self.caller.key} used the quit command")

            # manually call move hooks
            self.room.msg_room(self.caller, f"|r{self.caller.key} gave up and was whisked away!|n")
            self.room.at_object_leave(self.caller, self.caller.home)
            self.caller.move_to(
                self.caller.home, quiet=True, move_hooks=False, move_type="teleport"
            )

            # back to menu
            run_evscaperoom_menu(self.caller)
        else:
            self.msg("|gYou're staying? That's the spirit!|n")


class CmdLook(CmdEvscapeRoom):
    """
    Look at the room, an object or the currently focused object

    Usage:
      look [obj]

    """

    key = "look"
    aliases = ["l", "ls"]
    obj1_search = None
    obj2_search = None

    def func(self):
        caller = self.caller
        target = self.obj1 or self.obj2 or self.focus or self.room
        # the at_look hook will in turn call return_appearance and
        # pass the 'unfocused' kwarg to it
        txt = caller.at_look(target, unfocused=(target and target != self.focus))
        self.room.msg_char(caller, txt, client_type="look")


class CmdWho(CmdEvscapeRoom, default_cmds.CmdWho):
    """
    List other players in the game.

    Usage:
      who
      who all

    Show who is in the room with you, or (with who all), who is online on the
    server as a whole.

    """

    key = "who"

    obj1_search = False
    obj2_search = False

    def func(self):
        caller = self.caller

        if self.args == "all":
            table = self.style_table("|wName", "|wRoom")
            sessions = SESSION_HANDLER.get_sessions()
            for session in sessions:
                puppet = session.get_puppet()
                if puppet:
                    location = puppet.location
                    locname = location.key if location else "(Outside somewhere)"
                    table.add_row(puppet, locname)
                else:
                    account = session.get_account()
                    table.add_row(account.get_display_name(caller), "(OOC)")

            txt = (
                f"|cPlayers active on this server|n:\n{table}\n"
                "(use 'who' to see only those in your room)"
            )

        else:
            chars = [
                f"{obj.get_display_name(caller)} - {obj.db.desc.strip()}"
                for obj in self.room.get_all_characters()
                if obj != caller
            ]
            chars = "\n".join([f"{caller.key} - {caller.db.desc.strip()} (you)"] + chars)
            txt = f"|cPlayers in this room (room-name '{self.room.name}')|n:\n  {chars}"
        caller.msg(txt)


class CmdSpeak(Command):
    """
    Perform an communication action.

    Usage:
      say <text>
      whisper
      shout

    """

    key = "say"
    aliases = [";", "shout", "whisper"]
    arg_regex = r"\w|\s|$"

    def func(self):
        args = self.args.strip()
        caller = self.caller
        action = self.cmdname
        action = "say" if action == ";" else action
        room = self.caller.location

        if not self.args:
            caller.msg(f"What do you want to {action}?")
            return
        if action == "shout":
            args = f"|c{args.upper()}|n"
        elif action == "whisper":
            args = f"|C({args})|n"
        else:
            args = f"|c{args}|n"

        message = f"~You ~{action}: {args}"

        if hasattr(room, "msg_room"):
            room.msg_room(caller, message)
            room.log(f"{action} by {caller.key}: {args}")


class CmdEmote(Command):
    """
    Perform a free-form emote. Use /me to
    include yourself in the emote and /name
    to include other objects or characters.
    Use "..." to enact speech.

    Usage:
        emote <emote>
        :<emote

    Example:
        emote /me smiles at /peter
        emote /me points to /box and /lever.

    """

    key = "emote"
    aliases = [":", "pose"]
    arg_regex = r"\w|\s|$"

    def you_replace(match):
        return match

    def room_replace(match):
        return match

    def func(self):
        emote = self.args.strip()

        if not emote:
            self.caller.msg('Usage: emote /me points to /door, saying "look over there!"')
            return

        speech_clr = "|c"
        obj_clr = "|y"
        self_clr = "|g"
        player_clr = "|b"
        add_period = not _RE_EMOTE_PROPER_END.search(emote)

        emote = _RE_EMOTE_SPEECH.sub(speech_clr + r"\1\2|n", emote)
        room = self.caller.location

        characters = room.get_all_characters()
        logged = False
        for target in characters:
            txt = []
            self_refer = False
            for part in _RE_EMOTE_NAME.split(emote):
                nameobj = None
                if part.startswith("/"):
                    name = part[1:]
                    if name == "me":
                        nameobj = self.caller
                        self_refer = True
                    else:
                        match = self.caller.search(name, quiet=True)
                        if len(match) == 1:
                            nameobj = match[0]
                if nameobj:
                    if target == nameobj:
                        part = f"{self_clr}{nameobj.get_display_name(target)}|n"
                    elif nameobj in characters:
                        part = f"{player_clr}{nameobj.get_display_name(target)}|n"
                    else:
                        part = f"{obj_clr}{nameobj.get_display_name(target)}|n"
                txt.append(part)
            if not self_refer:
                if target == self.caller:
                    txt = [f"{self_clr}{self.caller.get_display_name(target)}|n "] + txt
                else:
                    txt = [f"{player_clr}{self.caller.get_display_name(target)}|n "] + txt
            txt = "".join(txt).strip() + ("." if add_period else "")
            if not logged and hasattr(self.caller.location, "log"):
                self.caller.location.log(f"emote: {txt}")
                logged = True
            target.msg(txt)


class CmdFocus(CmdEvscapeRoom):
    """
    Focus your attention on a target.

    Usage:
      focus <obj>

    Once focusing on an object, use look to get more information about how it
    looks and what actions is available.

    """

    key = "focus"
    aliases = ["examine", "e", "ex", "unfocus"]

    obj1_search = None

    def func(self):
        if self.obj1:
            old_focus = self.focus
            if hasattr(old_focus, "at_unfocus"):
                old_focus.at_unfocus(self.caller)

            if not hasattr(self.obj1, "at_focus"):
                self.caller.msg("Nothing of interest there.")
                return

            if self.focus != self.obj1:
                self.room.msg_room(
                    self.caller, f"~You ~examine *{self.obj1.key}.", skip_caller=True
                )
            self.focus = self.obj1
            self.obj1.at_focus(self.caller)
        elif not self.focus:
            self.caller.msg("What do you want to focus on?")
        else:
            old_focus = self.focus
            del self.focus
            self.caller.msg(f"You no longer focus on |y{old_focus.key}|n.")


class CmdOptions(CmdEvscapeRoom):
    """
    Start option menu

    Usage:
      options

    """

    key = "options"
    aliases = ["option"]

    def func(self):
        from .menu import run_option_menu

        run_option_menu(self.caller, self.session)


class CmdGet(CmdEvscapeRoom):
    """
    Use focus / examine instead.

    """

    key = "get"
    aliases = ["inventory", "i", "inv", "give"]

    def func(self):
        self.caller.msg("Use |wfocus|n or |wexamine|n for handling objects.")


class CmdRerouter(default_cmds.MuxCommand):
    """
    Interact with an object in focus.

    Usage:
       <action> [arg]

    """

    # reroute commands from the default cmdset to the catch-all
    # focus function where needed. This allows us to override
    # individual default commands without replacing the entire
    # cmdset (we want to keep most of them).

    key = "open"
    aliases = ["@dig", "@open"]

    def func(self):
        # reroute to another command
        from evennia.commands import cmdhandler

        cmdhandler.cmdhandler(
            self.session, self.raw_string, cmdobj=CmdFocusInteraction(), cmdobj_key=self.cmdname
        )


class CmdFocusInteraction(CmdEvscapeRoom):
    """
    Interact with an object in focus.

    Usage:
       <action> [arg]

    This is a special catch-all command which will operate on
    the current focus. It will look for a method
        `focused_object.at_focus_<action>(caller, **kwargs)` and call
    it. This allows objects to just add a new hook to make that
    action apply to it. The obj1, prep, obj2, arg1, arg2 are passed
    as keys into the method.

    """

    # all commands not matching something else goes here.
    key = syscmdkeys.CMD_NOMATCH

    obj1_search = None
    obj2_search = None

    def parse(self):
        """
        We assume this type of command is always on the form `command [arg]`

        """
        self.args = self.args.strip()
        parts = self.args.split(None, 1)
        if not self.args:
            self.action, self.args = "", ""
        elif len(parts) == 1:
            self.action = parts[0]
            self.args = ""
        else:
            self.action, self.args = parts
        self.room = self.caller.location

    def func(self):
        focused = self.focus
        action = self.action

        if focused and hasattr(focused, f"at_focus_{action}"):
            # there is a suitable hook to call!
            getattr(focused, f"at_focus_{action}")(self.caller, args=self.args)
        else:
            self.caller.msg("Hm?")


class CmdStand(CmdEvscapeRoom):
    """
    Stand up from whatever position you had.

    """

    key = "stand"

    def func(self):
        # Positionable objects will set this flag on you.
        pos = self.caller.attributes.get("position", category=self.room.tagcategory)

        if pos:
            # we have a position, clean up.
            obj, position = pos
            self.caller.attributes.remove("position", category=self.room.tagcategory)
            del obj.db.positions[self.caller]
            self.room.msg_room(self.caller, "~You ~are back standing on the floor again.")
        else:
            self.caller.msg("You are already standing.")


class CmdHelp(CmdEvscapeRoom, default_cmds.CmdHelp):
    """
    Get help.

    Usage:
      help <topic> or <command>

    """

    key = "help"
    aliases = ["?"]

    def func(self):
        if self.obj1:
            if hasattr(self.obj1, "get_help"):
                helptxt = self.obj1.get_help(self.caller)
                if not helptxt:
                    helptxt = f"There is no help to be had about {self.obj1.get_display_name(self.caller)}."
            else:
                helptxt = (
                    f"|y{self.obj1.get_display_name(self.caller)}|n is "
                    "likely |rnot|n part of any of the Jester's trickery."
                )
        elif self.arg1:
            # fall back to the normal help command
            super().func()
            return
        else:
            helptxt = _HELP_SUMMARY_TEXT
        self.caller.msg(helptxt.rstrip())


# Debug/help command


class CmdCreateObj(CmdEvscapeRoom):
    """
    Create command, only for Admins during debugging.

    Usage:
      createobj name[:typeclass]

    Here, :typeclass is a class in evscaperoom.commands

    """

    key = "createobj"
    aliases = ["cobj"]
    locks = "cmd:perm(Admin)"

    obj1_search = False
    obj2_search = False

    def func(self):
        caller = self.caller
        args = self.args

        if not args:
            caller.msg("Usage: createobj name[:typeclass]")
            return

        typeclass = "EvscaperoomObject"
        if ":" in args:
            name, typeclass = (part.strip() for part in args.rsplit(":", 1))

        if typeclass.startswith("state_"):
            # a state class
            typeclass = "evscaperoom.states." + typeclass
        else:
            name = args.strip()

        obj = create_evscaperoom_object(typeclass=typeclass, key=name, location=self.room)
        caller.msg(f"Created new object {name} ({obj.typeclass_path}).")


class CmdSetFlag(CmdEvscapeRoom):
    """
    Assign a flag to an object. Admin use only

    Usage:
      flag <obj> with <flagname>

    """

    key = "flag"
    aliases = ["setflag"]
    locks = "cmd:perm(Admin)"

    obj1_search = True
    obj2_search = False

    def func(self):
        if not self.arg2:
            self.caller.msg("Usage: flag <obj> with <flagname>")
            return

        if hasattr(self.obj1, "set_flag"):
            if self.obj1.check_flag(self.arg2):
                self.obj1.unset_flag(self.arg2)
                self.caller.msg(f"|rUnset|n flag '{self.arg2}' on {self.obj1}.")
            else:
                self.obj1.set_flag(self.arg2)
                self.caller.msg(f"|gSet|n flag '{self.arg2}' on {self.obj1}.")
        else:
            self.caller.msg(f"Cannot set flag on {self.obj1}.")


class CmdJumpState(CmdEvscapeRoom):
    """
    Jump to a given state.

    Args:
      jumpstate <statename>

    """

    key = "jumpstate"
    locks = "cmd:perm(Admin)"

    obj1_search = False
    obj2_search = False

    def func(self):
        self.caller.msg(f"Trying to move to state {self.args}")
        self.room.next_state(self.args)


# Helper command to start the Evscaperoom menu


class CmdEvscapeRoomStart(Command):
    """
    Go to the Evscaperoom start menu

    """

    key = "evscaperoom"
    help_category = "EvscapeRoom"

    def func(self):
        # need to import here to break circular import
        from .menu import run_evscaperoom_menu

        run_evscaperoom_menu(self.caller)


# command sets


class CmdSetEvScapeRoom(CmdSet):
    priority = 1

    def at_cmdset_creation(self):
        self.add(CmdHelp())
        self.add(CmdLook())
        self.add(CmdGiveUp())
        self.add(CmdFocus())
        self.add(CmdSpeak())
        self.add(CmdEmote())
        self.add(CmdFocusInteraction())
        self.add(CmdStand())
        self.add(CmdWho())
        self.add(CmdOptions())
        # rerouters
        self.add(CmdGet())
        self.add(CmdRerouter())
        # admin commands
        self.add(CmdCreateObj())
        self.add(CmdSetFlag())
        self.add(CmdJumpState())



"""
Command sets

All commands in the game must be grouped in a cmdset.  A given command
can be part of any number of cmdsets and cmdsets can be added/removed
and merged onto entities at runtime.

To create new commands to populate the cmdset, see
`commands/command.py`.

This module wraps the default command sets of Evennia; overloads them
to add/remove commands from the default lineup. You can create your
own cmdsets by inheriting from them or directly from `evennia.CmdSet`.

"""

from evennia import default_cmds


class CharacterCmdSet(default_cmds.CharacterCmdSet):
    """
    The `CharacterCmdSet` contains general in-game commands like `look`,
    `get`, etc available on in-game Character objects. It is merged with
    the `AccountCmdSet` when an Account puppets a Character.
    """

    key = "DefaultCharacter"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        super().at_cmdset_creation()
        #
        # any commands you add below will overload the default ones.
        #


class AccountCmdSet(default_cmds.AccountCmdSet):
    """
    This is the cmdset available to the Account at all times. It is
    combined with the `CharacterCmdSet` when the Account puppets a
    Character. It holds game-account-specific commands, channel
    commands, etc.
    """

    key = "DefaultAccount"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        super().at_cmdset_creation()
        #
        # any commands you add below will overload the default ones.
        #


class UnloggedinCmdSet(default_cmds.UnloggedinCmdSet):
    """
    Command set available to the Session before being logged in.  This
    holds commands like creating a new account, logging in, etc.
    """

    key = "DefaultUnloggedin"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        super().at_cmdset_creation()
        #
        # any commands you add below will overload the default ones.
        #


class SessionCmdSet(default_cmds.SessionCmdSet):
    """
    This cmdset is made available on Session level once logged in. It
    is empty by default.
    """

    key = "DefaultSession"

    def at_cmdset_creation(self):
        """
        This is the only method defined in a cmdset, called during
        its creation. It should populate the set with command instances.

        As and example we just add the empty base `Command` object.
        It prints some info.
        """
        super().at_cmdset_creation()
        #
        # any commands you add below will overload the default ones.
        #



"""
Unit tests for the LLM Client and npc.

"""

from anything import Something
from django.test import override_settings
from mock import Mock, patch

from evennia.utils.create import create_object
from evennia.utils.test_resources import BaseEvenniaTestCase

from .llm_npc import LLMNPC


class TestLLMClient(BaseEvenniaTestCase):
    """
    Test the LLMNPC class.

    """

    def setUp(self):
        self.npc = create_object(LLMNPC, key="Test NPC")
        self.npc.db_home = None  # fix a bug in test suite
        self.npc.save()

    def tearDown(self):
        self.npc.delete()
        super().tearDown()

    @override_settings(LLM_PROMPT_PREFIX="You are a test bot.")
    @patch("evennia.contrib.rpg.llm.llm_npc.task.deferLater")
    def test_npc_at_talked_to(self, mock_deferLater):
        """
        Test the npc's at_talked_to method.
        """
        mock_LLMClient = Mock()
        self.npc.ndb.llm_client = mock_LLMClient

        self.npc.at_talked_to("Hello", self.npc)

        mock_deferLater.assert_called_with(Something, self.npc.thinking_timeout, Something)
        mock_LLMClient.get_response.assert_called_with("You are a test bot.\nTest NPC: Hello")



import unittest

from django.conf import settings
from django.test import override_settings

from evennia import DefaultScript
from evennia.utils import containers
from evennia.utils.utils import class_from_module

_BASE_TYPECLASS = class_from_module(settings.BASE_SCRIPT_TYPECLASS)


class GoodScript(DefaultScript):
    pass


class InvalidScript:
    pass


class TestGlobalScriptContainer(unittest.TestCase):
    def test_init_with_no_scripts(self):
        gsc = containers.GlobalScriptContainer()

        self.assertEqual(len(gsc.loaded_data), 0)

    @override_settings(GLOBAL_SCRIPTS={})
    def test_start_with_no_scripts(self):
        gsc = containers.GlobalScriptContainer()

        gsc.start()

        self.assertEqual(len(gsc.typeclass_storage), 0)

    @override_settings(GLOBAL_SCRIPTS={"script_name": {}})
    def test_start_with_typeclassless_script(self):
        """No specified typeclass should fallback to base"""
        gsc = containers.GlobalScriptContainer()

        gsc.start()

        self.assertEqual(len(gsc.typeclass_storage), 1)
        self.assertIn("script_name", gsc.typeclass_storage)
        self.assertEqual(gsc.typeclass_storage["script_name"], _BASE_TYPECLASS)

    @override_settings(
        GLOBAL_SCRIPTS={
            "script_name": {"typeclass": "evennia.utils.tests.test_containers.NoScript"}
        }
    )
    def test_start_with_nonexistent_script(self):
        """Missing script class should fall back to base"""
        gsc = containers.GlobalScriptContainer()

        gsc.start()

        self.assertEqual(len(gsc.typeclass_storage), 1)
        self.assertIn("script_name", gsc.typeclass_storage)
        self.assertEqual(gsc.typeclass_storage["script_name"], _BASE_TYPECLASS)

    @override_settings(
        GLOBAL_SCRIPTS={
            "script_name": {"typeclass": "evennia.utils.tests.test_containers.GoodScript"}
        }
    )
    def test_start_with_valid_script(self):
        gsc = containers.GlobalScriptContainer()

        gsc.start()

        self.assertEqual(len(gsc.typeclass_storage), 1)
        self.assertIn("script_name", gsc.typeclass_storage)
        self.assertEqual(gsc.typeclass_storage["script_name"], GoodScript)

    @override_settings(
        GLOBAL_SCRIPTS={
            "script_name": {"typeclass": "evennia.utils.tests.test_containers.InvalidScript"}
        }
    )
    def test_start_with_invalid_script(self):
        """Script class doesn't implement required methods methods"""
        gsc = containers.GlobalScriptContainer()

        with self.assertRaises(AttributeError) as err:
            gsc.start()
        # check for general attribute failure on the invalid class to preserve against future code-rder changes
        self.assertTrue(
            str(err.exception).startswith("type object 'InvalidScript' has no attribute"),
            err.exception,
        )

    @override_settings(
        GLOBAL_SCRIPTS={
            "script_name": {"typeclass": "evennia.utils.tests.data.broken_script.BrokenScript"}
        }
    )
    def test_start_with_broken_script(self):
        """Un-importable script should traceback"""
        gsc = containers.GlobalScriptContainer()

        with self.assertRaises(Exception) as err:
            gsc.start()
        # exception raised by imported module
        self.assertTrue(
            str(err.exception).startswith("cannot import name 'nonexistent_module' from 'evennia'"),
            err.exception,
        )



"""
Traits

Whitenoise 2014, Ainneve contributors,
Griatch 2020


A `Trait` represents a modifiable property on (usually) a Character. They can
be used to represent everything from attributes (str, agi etc) to skills
(hunting 10, swords 14 etc) and dynamically changing things like HP, XP etc.

Traits use Evennia Attributes under the hood, making them persistent (they survive
a server reload/reboot).

## Installation

Traits are always added to a typeclass, such as the Character class.

There are two ways to set up Traits on a typeclass. The first sets up the `TraitHandler`
as a property `.traits` on your class and you then access traits as e.g. `.traits.strength`.
The other alternative uses a `TraitProperty`, which makes the trait available directly
as e.g. `.strength`. This solution also uses the `TraitHandler`, but you don't need to
define it explicitly. You can combine both styles if you like.

### Traits with TraitHandler

Here's an example for adding the TraitHandler to the Character class:

```python
# mygame/typeclasses/objects.py

from evennia import DefaultCharacter
from evennia.utils import lazy_property
from evennia.contrib.rpg.traits import TraitHandler

# ...

class Character(DefaultCharacter):
    ...
    @lazy_property
    def traits(self):
        # this adds the handler as .traits
        return TraitHandler(self)


    def at_object_creation(self):
        # (or wherever you want)
        self.traits.add("str", "Strength", trait_type="static", base=10, mod=2, mult=2.0)
        self.traits.add("hp", "Health", trait_type="gauge", min=0, max=100)
        self.traits.add("hunting", "Hunting Skill", trait_type="counter",
                        base=10, mod=1, min=0, max=100)


```
When adding the trait, you supply the name of the property (`hunting`) along
with a more human-friendly name ("Hunting Skill"). The latter will show if you
print the trait etc. The `trait_type` is important, this specifies which type
of trait this is (see below).

### TraitProperties

Using `TraitProperties` makes the trait available directly on the class, much like Django model
fields. The drawback is that you must make sure that the name of your Traits don't collide with any
other properties/methods on your class.

```python
# mygame/typeclasses/objects.py

from evennia import DefaultObject
from evennia.utils import lazy_property
from evennia.contrib.rpg.traits import TraitProperty

# ...

class Object(DefaultObject):
    ...
    strength = TraitProperty("Strength", trait_type="static", base=10, mod=2, mult=1.5)
    health = TraitProperty("Health", trait_type="gauge", min=0, base=100, mod=2)
    hunting = TraitProperty("Hunting Skill", trait_type="counter", base=10, mod=1, mult=2.0, min=0, max=100)

```

> Note that the property-name will become the name of the trait and you don't supply `trait_key`
> separately.

> The `.traits` TraitHandler will still be created (it's used under the
> hood. But it will only be created when the TraitProperty has been accessed at least once,
> so be careful if mixing the two styles. If you want to make sure `.traits` is always available,
> add the `TraitHandler` manually like shown earlier - the `TraitProperty` will by default use
> the same handler (`.traits`).

## Using traits

A trait is added to the traithandler (if you use `TraitProperty` the handler is just created under
the hood) after which one can access it as a property on the handler (similarly to how you can do
.db.attrname for Attributes in Evennia).

All traits have a _read-only_ field `.value`. This is only used to read out results, you never
manipulate it directly (if you try, it will just remain unchanged). The `.value` is calculated based
on combining fields, like `.base` and `.mod` - which fields are available and how they relate to
each other depends on the trait type.

```python
> obj.traits.strength.value
18                                  # (base + mod) * mult

> obj.traits.strength.base += 6
obj.traits.strength.value
27

> obj.traits.hp.value
102                                 # (base + mod) * mult

> obj.traits.hp.base -= 200
> obj.traits.hp.value
0                                   # min of 0

> obj.traits.hp.reset()
> obj.traits.hp.value
100

# you can also access properties like a dict
> obj.traits.hp["value"]
100

# you can store arbitrary data persistently for easy reference
> obj.traits.hp.effect = "poisoned!"
> obj.traits.hp.effect
"poisoned!"

# with TraitProperties:

> obj.hunting.value
22

> obj.strength.value += 5
> obj.strength.value
32

```

## Trait types

All default traits have a read-only `.value` property that shows the relevant or
'current' value of the trait. Exactly what this means depends on the type of trait.

Traits can also be combined to do arithmetic with their .value, if both have a
compatible type.

```python
> trait1 + trait2
54

> trait1.value
3

> trait1 + 2
> trait1.value
5

```

Two numerical traits can also be compared (bigger-than etc), which is useful in
all sorts of rule-resolution.

```python

if trait1 > trait2:
    # do stuff

```
## Static trait

`value = (base + mod) * mult`

The static trait has a `base` value and an optional `mod`-ifier and 'mult'-iplier.
The modifier defaults to 0, and the multiplier to 1.0, for no change in value.
A typical use of a static trait would be a Strength stat or Skill value. That is,
somethingthat varies slowly or not at all, and which may be modified in-place.

```python
> obj.traits.add("str", "Strength", trait_type="static", base=10, mod=2)
> obj.traits.mytrait.value
12   # base + mod

> obj.traits.mytrait.base += 2
> obj.traits.mytrait.mod += 1
> obj.traits.mytrait.value
15

> obj.traits.mytrait.mod = 0
> obj.traits.mytrait.mult = 2.0
> obj.traits.mytrait.value
20

```

### Counter
::

    min/unset     base    base+mod                       max/unset
    |--------------|--------|---------X--------X------------|
                                  current    value
                                             = (current
                                             + mod)
                                             * mult

A counter describes a value that can move from a base. The `.current` property
is the thing usually modified. It starts at the `.base`. One can also add a
modifier, which is added to both the base and to current. '.value' is then formed
by multiplying by the multiplier, which defaults to 1.0 for no change. The min/max
of the range are optional, a boundary set to None will remove it. A suggested use
for a Counter Trait would be to track skill values.

```python
> obj.traits.add("hunting", "Hunting Skill", trait_type="counter",
                   base=10, mod=1, mult=1.0, min=0, max=100)
> obj.traits.hunting.value
11  # current starts at base + mod

> obj.traits.hunting.current += 10
> obj.traits.hunting.value
21

# reset back to base+mod by deleting current
> del obj.traits.hunting.current
> obj.traits.hunting.value
11

> obj.traits.hunting.max = None  # removing upper bound
> obj.traits.hunting.mult = 100.0
1100

# for TraitProperties, pass the args/kwargs of traits.add() to the
# TraitProperty constructor instead.


```

Counters have some extra properties:

#### .descs

The `descs` property is a dict {upper_bound:text_description}. This allows for easily
storing a more human-friendly description of the current value in the
interval. Here is an example for skill values between 0 and 10:
::

    {0: "unskilled", 1: "neophyte", 5: "trained", 7: "expert", 9: "master"}

The keys must be supplied from smallest to largest. Any values below the lowest and above the
highest description will be considered to be included in the closest description slot.
By calling `.desc()` on the Counter, you will get the text matching the current `value`.

```python
# (could also have passed descs= to traits.add())
> obj.traits.hunting.descs = {
    0: "unskilled", 10: "neophyte", 50: "trained", 70: "expert", 90: "master"}
> obj.traits.hunting.value
11

> obj.traits.hunting.desc()
"neophyte"
> obj.traits.hunting.current += 60
> obj.traits.hunting.value
71

> obj.traits.hunting.desc()
"expert"

```

#### .rate

The `rate` property defaults to 0. If set to a value different from 0, it
allows the trait to change value dynamically. This could be used for example
for an attribute that was temporarily lowered but will gradually (or abruptly)
recover after a certain time. The rate is given as change of the current
`.value` per-second, and this will still be restrained by min/max boundaries,
if those are set.

It is also possible to set a `.ratetarget`, for the auto-change to stop at
(rather than at the min/max boundaries). This allows the value to return to
a previous value.

```python

> obj.traits.hunting.value
71

> obj.traits.hunting.ratetarget = 71
# debuff hunting for some reason
> obj.traits.hunting.current -= 30
> obj.traits.hunting.value
41

> obj.traits.hunting.rate = 1  # 1/s increase
# Waiting 5s
> obj.traits.hunting.value
46

# Waiting 8s
> obj.traits.hunting.value
54

# Waiting 100s
> obj.traits.hunting.value
71    # we have stopped at the ratetarget

> obj.traits.hunting.rate = 0  # disable auto-change


```
Note that when retrieving the `current`, the result will always be of the same
type as the `.base` even `rate` is a non-integer value. So if `base` is an `int`
(default)`, the `current` value will also be rounded the closest full integer.
If you want to see the exact `current` value, set `base` to a float - you
will then need to use `round()` yourself on the result if you want integers.

#### .percent()

If both min and max are defined, the `.percent()` method of the trait will
return the value as a percentage.

```python
> obj.traits.hunting.percent()
"71.0%"

> obj.traits.hunting.percent(formatting=None)
71.0

```

### Gauge

This emulates a [fuel-] gauge that empties from a base+mod value.
::

    min/0                                            max=base+mod
     |-----------------------X---------------------------|
                           value
                          = current

The `.current` value will start from a full gauge. The .max property is
read-only and is set by `.base` + `.mod`. So contrary to a `Counter`, the
`.mod` modifier only applies to the max value of the gauge and not the current
value. The minimum bound defaults to 0 if not set explicitly.

This trait is useful for showing commonly depletable resources like health,
stamina and the like.

```python
> obj.traits.add("hp", "Health", trait_type="gauge", base=100)
> obj.traits.hp.value  # (or .current)
100

> obj.traits.hp.mod = 10
> obj.traits.hp.value
110

> obj.traits.hp.current -= 30
> obj.traits.hp.value
80

```

The Gauge trait is subclass of the Counter, so you have access to the same
methods and properties where they make sense. So gauges can also have a
`.descs` dict to describe the intervals in text, and can use `.percent()` to
get how filled it is as a percentage etc.

The `.rate` is particularly relevant for gauges - useful for everything
from poison slowly draining your health, to resting gradually increasing it.

### Trait

A single value of any type.

This is the 'base' Trait, meant to inherit from if you want to invent
trait-types from scratch (most of the time you'll probably inherit from some of
the more advanced trait-type classes though).

Unlike other Trait-types, the single `.value` property of the base `Trait` can
be editied. The value can hold any data that can be stored in an Attribute. If
it's an integer/float you can do arithmetic with it, but otherwise this acts just
like a glorified Attribute.


```python
> obj.traits.add("mytrait", "My Trait", trait_type="trait", value=30)
> obj.traits.mytrait.value
30

> obj.traits.mytrait.value = "stringvalue"
> obj.traits.mytrait.value
"stringvalue"

```

## Expanding with your own Traits

A Trait is a class inhering from `evennia.contrib.rpg.traits.Trait` (or from one of
the existing Trait classes).

```python
# in a file, say, 'mygame/world/traits.py'

from evennia.contrib.rpg.traits import StaticTrait

class RageTrait(StaticTrait):

    trait_type = "rage"
    default_keys = {
        "rage": 0
    }

    def berserk(self):
        self.mod = 100

    def sedate(self):
        self.mod = 0


```

Above is an example custom-trait-class "rage" that stores a property "rage" on
itself, with a default value of 0. This has all the functionality of a Trait -
for example, if you do del on the `rage` property, it will be set back to its
default (0). Above we also added some helper methods.

To add your custom RageTrait to Evennia, add the following to your settings file
(assuming your class is in mygame/world/traits.py):
::

    TRAIT_CLASS_PATHS = ["world.traits.RageTrait"]

Reload the server and you should now be able to use your trait:

```python
> obj.traits.add("mood", "A dark mood", rage=30, trait_type='rage')
> obj.traits.mood.rage
30

# as TraitProperty

class Character(DefaultCharacter):
    rage = TraitProperty("A dark mood", rage=30, trait_type='rage')

```

----

"""


from functools import total_ordering
from time import time

from django.conf import settings

from evennia.utils import logger
from evennia.utils.dbserialize import _SaverDict
from evennia.utils.utils import (
    class_from_module,
    inherits_from,
    list_to_string,
    percent,
)

# Available Trait classes.
# This way the user can easily supply their own. Each
# class should have a class-property `trait_type` to
# identify the Trait class. The default ones are "static",
# "counter" and "gauge".

_TRAIT_CLASS_PATHS = [
    "evennia.contrib.rpg.traits.Trait",
    "evennia.contrib.rpg.traits.StaticTrait",
    "evennia.contrib.rpg.traits.CounterTrait",
    "evennia.contrib.rpg.traits.GaugeTrait",
]

if hasattr(settings, "TRAIT_CLASS_PATHS"):
    _TRAIT_CLASS_PATHS += settings.TRAIT_CLASS_PATHS

# delay trait-class import to avoid circular import
_TRAIT_CLASSES = None


def _delayed_import_trait_classes():
    """
    Import classes based on the given paths. Note that
    imports from settings are last in the list, so if they
    have the same trait_type set, they will replace the
    default.
    """
    global _TRAIT_CLASSES
    if _TRAIT_CLASSES is None:
        _TRAIT_CLASSES = {}
        for classpath in _TRAIT_CLASS_PATHS:
            try:
                cls = class_from_module(classpath)
            except ImportError:
                logger.log_trace(f"Could not import Trait from {classpath}.")
            else:
                if hasattr(cls, "trait_type"):
                    trait_type = cls.trait_type
                else:
                    trait_type = str(cls.__name___).lower()
                _TRAIT_CLASSES[trait_type] = cls


_GA = object.__getattribute__
_SA = object.__setattr__
_DA = object.__delattr__

# this is the default we offer in TraitHandler.add
DEFAULT_TRAIT_TYPE = "static"


class TraitException(RuntimeError):
    """
    Base exception class raised by `Trait` objects.

    Args:
        msg (str): informative error message

    """

    def __init__(self, msg):
        self.msg = msg


class MandatoryTraitKey:
    """
    This represents a required key that must be
    supplied when a Trait is initialized. It's used
    by Trait classes when defining their required keys.

    """


class TraitHandler:
    """
    Factory class that instantiates Trait objects. Must be assigned as a property
    on the class, usually with `lazy_property`.

    Example:
    ::
        class Object(DefaultObject):
            ...
            @lazy_property
            def traits(self):
                # this adds the handler as .traits
                return TraitHandler(self)

    """

    def __init__(self, obj, db_attribute_key="traits", db_attribute_category="traits"):
        """
        Initialize the handler and set up its internal Attribute-based storage.

        Args:
            obj (Object): Parent Object typeclass for this TraitHandler
            db_attribute_key (str): Name of the DB attribute for trait data storage.
            db_attribute_category (str):  Name of DB attribute's category to trait data storage.

        """
        # load the available classes, if necessary
        _delayed_import_trait_classes()

        # initialize any
        # Note that .trait_data retains the connection to the database, meaning every
        # update we do to .trait_data automatically syncs with database.
        self.trait_data = obj.attributes.get(db_attribute_key, category=db_attribute_category)
        if self.trait_data is None:
            # no existing storage; initialize it, we then have to fetch it again
            # to retain the db connection
            obj.attributes.add(db_attribute_key, {}, category=db_attribute_category)
            self.trait_data = obj.attributes.get(db_attribute_key, category=db_attribute_category)
        self._cache = {}

    def __len__(self):
        """Return number of Traits registered with the handler"""
        return len(self.trait_data)

    def __setattr__(self, trait_key, value):
        """
        Returns error message if trait objects are assigned directly.

        Args:
            trait_key (str): The Trait-key, like "hp".
            value (any): Data to store.
        """
        if trait_key in ("trait_data", "_cache"):
            _SA(self, trait_key, value)
        else:
            trait_cls = self._get_trait_class(trait_key=trait_key)
            valid_keys = list_to_string(list(trait_cls.default_keys.keys()), endsep="or")
            raise TraitException(
                f"Trait object not settable directly. Assign to {trait_key}.{valid_keys}."
            )

    def __setitem__(self, trait_key, value):
        """Returns error message if trait objects are assigned directly."""
        return self.__setattr__(trait_key, value)

    def __getattr__(self, trait_key):
        """Returns Trait instances accessed as attributes."""
        return self.get(trait_key)

    def __getitem__(self, trait_key):
        """Returns `Trait` instances accessed as dict keys."""
        return self.get(trait_key)

    def __repr__(self):
        return "TraitHandler ({num} Trait(s) stored): {keys}".format(
            num=len(self), keys=", ".join(self.all())
        )

    def _get_trait_class(self, trait_type=None, trait_key=None):
        """
        Helper to retrieve Trait class based on type (like "static")
        or trait-key (like "hp").

        """
        if not trait_type and trait_key:
            try:
                trait_type = self.trait_data[trait_key]["trait_type"]
            except KeyError:
                raise TraitException(f"Trait class for Trait {trait_key} could not be found.")
        try:
            return _TRAIT_CLASSES[trait_type]
        except KeyError:
            raise TraitException(f"Trait class for {trait_type} could not be found.")

    def all(self):
        """
        Get all trait keys in this handler.

        Returns:
            list: All Trait keys.

        """
        return list(self.trait_data.keys())

    def get(self, trait_key):
        """
        Args:
            trait_key (str): key from the traits dict containing config data.

        Returns:
            (`Trait` or `None`): named Trait class or None if trait key
            is not found in traits collection.

        """
        trait = self._cache.get(trait_key)
        if trait is None and trait_key in self.trait_data:
            trait_type = self.trait_data[trait_key]["trait_type"]
            trait_cls = self._get_trait_class(trait_type)
            trait = self._cache[trait_key] = trait_cls(_GA(self, "trait_data")[trait_key])
        return trait

    def add(
        self, trait_key, name=None, trait_type=DEFAULT_TRAIT_TYPE, force=True, **trait_properties
    ):
        """
        Create a new Trait and add it to the handler.

        Args:
            trait_key (str): This is the name of the property that will be made
                available on this handler (example 'hp').
            name (str, optional): Name of the Trait, like "Health". If
                not given, will use `trait_key` starting with a capital letter.
            trait_type (str, optional): One of 'static', 'counter' or 'gauge'.
            force (bool): If set, create a new Trait even if a Trait with
                the same `trait_key` already exists.
            trait_properties (dict): These will all be use to initialize
                the new trait. See the `properties` class variable on each
                Trait class to see which are required.

        Raises:
            TraitException: If specifying invalid values for the given Trait,
                the `trait_type` is not recognized, or an existing trait
                already exists (and `force` is unset).

        """
        # from evennia import set_trace;set_trace()

        if trait_key in self.trait_data:
            if force:
                self.remove(trait_key)
            else:
                raise TraitException(f"Trait '{trait_key}' already exists.")

        trait_class = _TRAIT_CLASSES.get(trait_type)
        if not trait_class:
            raise TraitException(f"Trait-type '{trait_type}' is invalid.")

        trait_properties["name"] = trait_key.title() if not name else name
        trait_properties["trait_type"] = trait_type

        # this will raise exception if input is insufficient
        trait_properties = trait_class.validate_input(trait_class, trait_properties)

        self.trait_data[trait_key] = trait_properties

    def remove(self, trait_key):
        """
        Remove a Trait from the handler's parent object.

        Args:
            trait_key (str): The name of the trait to remove.

        """
        if trait_key not in self.trait_data:
            raise TraitException(f"Trait '{trait_key}' not found.")

        if trait_key in self._cache:
            del self._cache[trait_key]
        del self.trait_data[trait_key]

    def clear(self):
        """
        Remove all Traits from the handler's parent object.
        """
        for trait_key in self.all():
            self.remove(trait_key)


class TraitProperty:
    """
    Optional extra: Allows for applying traits as individual properties directly on the parent class
    instead for properties on the `.traits` handler. So with this you could access data e.g. as
    `character.hp.value` instead of `character.traits.hp.value`. This still uses the traitshandler
    under the hood.

    Example:
    ::
        from evennia.utils import lazy_property
        from evennia.contrib.rpg.traits import TraitProperty

        class Character(DefaultCharacter):

            strength = TraitProperty(name="STR", trait_type="static", base=10, mod=2)
            hunting = TraitProperty("Hunting Skill", trait_type="counter",
                                    base=10, mod=1, max=100)
            health = TraitProperty(trait_type="gauge", min=0, base=100)

    """

    def __init__(self, name=None, trait_type=DEFAULT_TRAIT_TYPE, force=True, **trait_properties):
        """
        Initialize a TraitField. Mimics TraitHandler.add input except no `trait_key`.

        Args:
            name (str, optional): Name of the Trait, like "Health". If
                not given, will use `trait_key` starting with a capital letter.
            trait_type (str, optional): One of 'static', 'counter' or 'gauge'.
            force (bool): If set, create a new Trait even if a Trait with
                the same `trait_key` already exists.
        Kwargs:
            traithandler_name (str): If given, this is used as the name of the TraitHandler created
                behind the scenes. If not set, this will be a property `traits` on the class.
            any: All other trait_properties are the same as for adding a new trait of the given type
                using the normal TraitHandler.

        """
        self._traithandler_name = trait_properties.pop("traithandler_name", "traits")

        trait_properties.update({"name": name, "trait_type": trait_type, "force": force})
        self._trait_properties = trait_properties
        self._cache = {}

    def __set_name__(self, instance, name):
        """
        This is called the very first time the Descriptor is assigned to the
        class; we store it so we can create new instances with this later.

        """
        self._trait_key = name

    def __get__(self, instance, owner):
        """
        Descriptor definition. This is called when the trait-name is aqcuired on the
        instance and reroutes to fetching the actual Trait from the connected
        TraitHandler (the connection is set up on-demand).

        Returns:
            Trait: The trait this property represents.

        Notes:
            We have one descriptor on the class, but we don't want each instance to share the
            state (self) of that descriptor. So we must make sure to cache the trait per-instance
            or we would end up with cross-use between instances.

        """
        if instance not in self._cache:
            try:
                traithandler = getattr(instance, self._traithandler_name)
            except AttributeError:
                # traithandler not found; create a new on-demand
                traithandler = TraitHandler(instance)
                setattr(instance, self._traithandler_name, traithandler)

            # this will either get the trait from attribute or make a new one
            trait = traithandler.get(self._trait_key)
            if trait is None:
                # initialize the trait
                traithandler.add(self._trait_key, **self._trait_properties)
                trait = traithandler.get(self._trait_key)  # caches it in the traithandler
            self._cache[instance] = trait
        return self._cache[instance]

    def __set__(self, instance, value):
        """
        We don't set data directly, it's all rerouted to the trait.

        """
        pass


# Parent Trait class


@total_ordering
class Trait:
    """Represents an object or Character trait. This simple base is just
    storing anything in it's 'value' property, so it's pretty much just a
    different wrapper to an Attribute. It does no type-checking of what is
    stored.

    Note:
        See module docstring for configuration details.

    value

    """

    # this is the name used to refer to this trait when adding
    # a new trait in the TraitHandler
    trait_type = "trait"

    # Property kwargs settable when creating a Trait of this type. This is a
    # dict of key: default. To indicate a mandatory kwarg and raise an error if
    # not given, set the default value to the `traits.MandatoryTraitKey` class.
    # Apart from the keys given here, "name" and "trait_type" will also always
    # have to be a apart of the data.
    default_keys = {"value": None}

    # enable to set/retrieve other arbitrary properties on the Trait
    # and have them treated like data to store.
    allow_extra_properties = True

    def __init__(self, trait_data):
        """
        This both initializes and validates the Trait on creation. It must
        raise exception if validation fails. The TraitHandler will call this
        when the trait is furst added, to make sure it validates before
        storing.

        Args:
            trait_data (any): Any pickle-able values to store with this trait.
                This must contain any cls.default_keys that do not have a default
                value in cls.data_default_values. Any extra kwargs will be made
                available as extra properties on the Trait, assuming the class
                variable `allow_extra_properties` is set.

        Raises:
            TraitException: If input-validation failed.

        """
        self._data = self.__class__.validate_input(self.__class__, trait_data)

        if not isinstance(trait_data, _SaverDict):
            logger.log_warn(
                f"Non-persistent Trait data (type(trait_data)) loaded for {type(self).__name__}."
            )

    @staticmethod
    def validate_input(cls, trait_data):
        """
        Validate input

        Args:
            trait_data (dict or _SaverDict): Data to be used for
                initialization of this trait.
        Returns:
            dict: Validated data, possibly complemented with default
                values from default_keys.
        Raises:
            TraitException: If finding unset keys without a default.

        """

        def _raise_err(unset_required):
            """Helper method to format exception."""
            raise TraitException(
                "Trait {} could not be created - misses required keys {}.".format(
                    cls.trait_type, list_to_string(list(unset_required), addquote=True)
                )
            )

        inp = set(trait_data.keys())

        # separate check for name/trait_type, those are always required.
        req = set(("name", "trait_type"))
        unsets = req.difference(inp.intersection(req))
        if unsets:
            _raise_err(unsets)

        # check other keys, these likely have defaults to fall back to
        req = set(list(cls.default_keys.keys()))
        unsets = req.difference(inp.intersection(req))
        unset_defaults = {key: cls.default_keys[key] for key in unsets}

        if MandatoryTraitKey in unset_defaults.values():
            # we have one or more unset keys that was mandatory
            _raise_err([key for key, value in unset_defaults.items() if value == MandatoryTraitKey])
        # apply the default values
        trait_data.update(unset_defaults)

        if not cls.allow_extra_properties:
            # don't allow any extra properties - remove the extra data
            for key in (key for key in inp.difference(req) if key not in ("name", "trait_type")):
                del trait_data[key]

        return trait_data

    # Grant access to properties on this Trait.

    def __getitem__(self, key):
        """Access extra parameters as dict keys."""
        try:
            return self.__getattr__(key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        """Set extra parameters as dict keys."""
        self.__setattr__(key, value)

    def __delitem__(self, key):
        """Delete extra parameters as dict keys."""
        self.__delattr__(key)

    def __getattr__(self, key):
        """Access extra parameters as attributes."""
        if key in ("default_keys", "data_default", "trait_type", "allow_extra_properties"):
            return _GA(self, key)
        try:
            return self._data[key]
        except KeyError:
            raise AttributeError(
                "{!r} {} ({}) has no property {!r}.".format(
                    self._data["name"], type(self).__name__, self.trait_type, key
                )
            )

    def __setattr__(self, key, value):
        """Set extra parameters as attributes.

        Arbitrary attributes set on a Trait object will be
        stored in the 'extra' key of the `_data` attribute.

        This behavior is enabled by setting the instance
        variable `_locked` to True.

        """
        propobj = getattr(self.__class__, key, None)
        if isinstance(propobj, property):
            # we have a custom property named as this key, find and use its setter
            if propobj.fset:
                propobj.fset(self, value)
            return
        else:
            # this is some other value
            if key in ("_data",):
                _SA(self, key, value)
                return
            if _GA(self, "allow_extra_properties"):
                _GA(self, "_data")[key] = value
                return
        raise AttributeError(f"Can't set attribute {key} on {self.trait_type} Trait.")

    def __delattr__(self, key):
        """
        Delete or reset parameters.

        Args:
            key (str): property-key to delete.
        Raises:
            TraitException: If trying to delete a data-key
                without a default value to reset to.
        Notes:
            This will outright delete extra keys (if allow_extra_properties is
            set). Keys in self.default_keys with a default value will be
            reset to default. A data_key with a default of MandatoryDefaultKey
            will raise a TraitException. Unfound matches will be silently ignored.

        """
        if key in self.default_keys:
            if self.default_keys[key] == MandatoryTraitKey:
                raise TraitException(
                    "Trait-Key {key} cannot be deleted: It's a mandatory property "
                    "with no default value to fall back to."
                )
            # set to default
            self._data[key] = self.default_keys[key]
        elif key in self._data:
            try:
                # check if we have a custom deleter
                _DA(self, key)
            except AttributeError:
                # delete normally
                del self._data[key]
        else:
            try:
                # check if we have custom deleter, otherwise ignore
                _DA(self, key)
            except AttributeError:
                pass

    def __repr__(self):
        """Debug-friendly representation of this Trait."""
        return "{}({{{}}})".format(
            type(self).__name__,
            ", ".join(
                [
                    "'{}': {!r}".format(k, self._data[k])
                    for k in self.default_keys
                    if k in self._data
                ]
            ),
        )

    def __str__(self):
        return f"<Trait {self.name}: {self._data['value']}>"

    # access properties

    @property
    def name(self):
        """Display name for the trait."""
        return self._data["name"]

    key = name

    # Numeric operations

    def __eq__(self, other):
        """Support equality comparison between Traits or Trait and numeric.

        Note:
            This class uses the @functools.total_ordering() decorator to
            complete the rich comparison implementation, therefore only
            `__eq__` and `__lt__` are implemented.
        """
        if inherits_from(other, Trait):
            return self.value == other.value
        elif type(other) in (float, int):
            return self.value == other
        else:
            return NotImplemented

    def __lt__(self, other):
        """Support less than comparison between `Trait`s or `Trait` and numeric."""
        if inherits_from(other, Trait):
            return self.value < other.value
        elif type(other) in (float, int):
            return self.value < other
        else:
            return NotImplemented

    def __pos__(self):
        """Access `value` property through unary `+` operator."""
        return self.value

    def __add__(self, other):
        """Support addition between `Trait`s or `Trait` and numeric"""
        if inherits_from(other, Trait):
            return self.value + other.value
        elif type(other) in (float, int):
            return self.value + other
        else:
            return NotImplemented

    def __sub__(self, other):
        """Support subtraction between `Trait`s or `Trait` and numeric"""
        if inherits_from(other, Trait):
            return self.value - other.value
        elif type(other) in (float, int):
            return self.value - other
        else:
            return NotImplemented

    def __mul__(self, other):
        """Support multiplication between `Trait`s or `Trait` and numeric"""
        if inherits_from(other, Trait):
            return self.value * other.value
        elif type(other) in (float, int):
            return self.value * other
        else:
            return NotImplemented

    def __floordiv__(self, other):
        """Support floor division between `Trait`s or `Trait` and numeric"""
        if inherits_from(other, Trait):
            return self.value // other.value
        elif type(other) in (float, int):
            return self.value // other
        else:
            return NotImplemented

    # commutative property
    __radd__ = __add__
    __rmul__ = __mul__

    def __rsub__(self, other):
        """Support subtraction between `Trait`s or `Trait` and numeric"""
        if inherits_from(other, Trait):
            return other.value - self.value
        elif type(other) in (float, int):
            return other - self.value
        else:
            return NotImplemented

    def __rfloordiv__(self, other):
        """Support floor division between `Trait`s or `Trait` and numeric"""
        if inherits_from(other, Trait):
            return other.value // self.value
        elif type(other) in (float, int):
            return other // self.value
        else:
            return NotImplemented

    # Public members

    @property
    def value(self):
        """Store a value"""
        return self._data["value"]

    @value.setter
    def value(self, value):
        """Get value"""
        self._data["value"] = value


# Implementation of the respective Trait types


class StaticTrait(Trait):
    """
    Static Trait. This is a single value with a modifier,
    multiplier, and no concept of a 'current' value or min/max etc.

    value = (base + mod) * mult

    """

    trait_type = "static"

    default_keys = {"base": 0, "mod": 0, "mult": 1.0}

    def __str__(self):
        status = "{value:11}".format(value=self.value)
        return "{name:12} {status} ({mod:+3}) (* {mult:.2f})".format(
            name=self.name, status=status, mod=self.mod, mult=self.mult
        )

    # Helpers
    @property
    def base(self):
        return self._data["base"]

    @base.setter
    def base(self, value):
        if value is None:
            self._data["base"] = self.default_keys["base"]
        if type(value) in (int, float):
            self._data["base"] = value

    @property
    def mod(self):
        """The trait's modifier."""
        return self._data["mod"]

    @mod.setter
    def mod(self, amount):
        if type(amount) in (int, float):
            self._data["mod"] = amount

    @property
    def mult(self):
        """The trait's multiplier."""
        return self._data["mult"]

    @mult.setter
    def mult(self, amount):
        if type(amount) in (int, float):
            self._data["mult"] = amount

    @mult.deleter
    def mult(self):
        self._data["mult"] = 1.0

    @property
    def value(self):
        "The value of the Trait."
        return (self.base + self.mod) * self.mult


class CounterTrait(Trait):
    """
    Counter Trait.

    This includes modifications and min/max limits as well as the notion of a
    current value. The value can also be reset to the base value.

    min/unset     base  (base+mod)*mult                  max/unset
     |--------------|--------|---------X--------X------------|
                                    current   value
                                              = (current
                                              + mod)
                                              * mult

    - value = (current + mod) * mult, starts at (base + mod) * mult
    - if min or max is None, there is no upper/lower bound (default)
    - if max is set to "base", max will be equal ot base+mod
    - descs are used to optionally describe each value interval.
      The desc of the current `value` value can then be retrieved
      with .desc(). The property is set as {lower_bound_inclusive:desc}
      and should be given smallest-to-biggest. For example, for
      a skill rating between 0 and 10:
            {0: "unskilled",
             1: "neophyte",
             5: "traited",
             7: "expert",
             9: "master"}
    - rate/ratetarget are optional settings to include a rate-of-change
      of the current value. This is calculated on-demand and allows for
      describing a value that is gradually growing smaller/bigger. The
      increase will stop when either reaching a boundary (if set) or
      ratetarget. Setting the rate to 0 (default) stops any change.

    """

    trait_type = "counter"

    # current starts equal to base.
    default_keys = {
        "base": 0,
        "mod": 0,
        "mult": 1.0,
        "min": None,
        "max": None,
        "descs": None,
        "rate": 0,
        "ratetarget": None,
    }

    @staticmethod
    def validate_input(cls, trait_data):
        """Add extra validation for descs"""
        trait_data = Trait.validate_input(cls, trait_data)
        # validate descs
        descs = trait_data["descs"]
        if isinstance(descs, dict):
            if any(
                not (isinstance(key, (int, float)) and isinstance(value, str))
                for key, value in descs.items()
            ):
                raise TraitException(
                    "Trait descs must be defined on the "
                    f"form {{number:str}} (instead found {descs})."
                )
        # set up rate
        if trait_data["rate"] != 0:
            trait_data["last_update"] = time()
        else:
            trait_data["last_update"] = None
        return trait_data

    def __str__(self):
        status = "{current:4} / {base:4}".format(current=self.current, base=self.base)
        return "{name:12} {status} ({mod:+3}) (* {mult:.2f})".format(
            name=self.name, status=status, mod=self.mod, mult=self.mult
        )

    # Helpers

    def _within_boundaries(self, value):
        """Check if given value is within boundaries"""
        return not (
            (self.min is not None and value <= self.min)
            or (self.max is not None and value >= self.max)
        )

    def _enforce_boundaries(self, value):
        """Ensures that incoming value falls within boundaries"""
        if self.min is not None and value <= self.min:
            return self.min
        if self.max is not None and value >= self.max:
            return self.max
        return value

    # timer component

    def _passed_ratetarget(self, value):
        """Check if we passed the ratetarget in either direction."""
        ratetarget = self._data["ratetarget"]
        return ratetarget is not None and (
            (self.rate < 0 and value <= ratetarget) or (self.rate > 0 and value >= ratetarget)
        )

    def _stop_timer(self):
        """Stop rate-timer component."""
        if self.rate != 0 and self._data["last_update"] is not None:
            self._data["last_update"] = None

    def _check_and_start_timer(self, value):
        """Start timer if we are not at a boundary."""
        if self.rate != 0 and self._data["last_update"] is None:
            if self._within_boundaries(value) and not self._passed_ratetarget(value):
                # we are not at a boundary [anymore].
                self._data["last_update"] = time()
        return value

    def _update_current(self, current):
        """Update current value by scaling with rate and time passed."""
        rate = self.rate
        if rate != 0 and self._data["last_update"] is not None:
            now = time()
            tdiff = now - self._data["last_update"]
            current += rate * tdiff
            value = current + self.mod

            # we must make sure so we don't overstep our bounds
            # even if .mod is included

            if self._passed_ratetarget(value):
                current = self._data["ratetarget"] - self.mod
                self._stop_timer()
            elif not self._within_boundaries(value):
                current = self._enforce_boundaries(value) - self.mod
                self._stop_timer()
            else:
                self._data["last_update"] = now

            self._data["current"] = current

        if self.base is not None and isinstance(self.base, int):
            return round(current)
        return current

    # properties

    @property
    def base(self):
        return self._data["base"]

    @base.setter
    def base(self, value):
        if value is None:
            self._data["base"] = self.default_keys["base"]
        if type(value) in (int, float):
            if self.min is not None and value + self.mod < self.min:
                value = self.min - self.mod
            if self.max is not None and value + self.mod > self.max:
                value = self.max - self.mod
            self._data["base"] = value

    @property
    def mod(self):
        return self._data["mod"]

    @mod.setter
    def mod(self, value):
        if value is None:
            # unsetting the boundary to default
            self._data["mod"] = self.default_keys["mod"]
        elif type(value) in (int, float):
            if self.min is not None and value + self.base < self.min:
                value = self.min - self.base
            if self.max is not None and value + self.base > self.max:
                value = self.max - self.base
            self._data["mod"] = value

    @property
    def mult(self):
        return self._data["mult"]

    @mult.setter
    def mult(self, amount):
        if type(amount) in (int, float):
            self._data["mult"] = amount

    @mult.deleter
    def mult(self):
        self._data["mult"] = 1.0

    @property
    def min(self):
        return self._data["min"]

    @min.setter
    def min(self, value):
        if value is None:
            # unsetting the boundary
            self._data["min"] = value
        elif type(value) in (int, float):
            if self.max is not None:
                value = min(self.max, value)
            self._data["min"] = min(value, self.base + self.mod)

    @property
    def max(self):
        return self._data["max"]

    @max.setter
    def max(self, value):
        if value is None:
            # unsetting the boundary
            self._data["max"] = value
        elif type(value) in (int, float):
            if self.min is not None:
                value = max(self.min, value)
            self._data["max"] = max(value, self.base + self.mod)

    @property
    def current(self):
        """The `current` value of the `Trait`. This does not have .mod added and is not .mult-iplied."""
        return self._update_current(self._data.get("current", self.base))

    @current.setter
    def current(self, value):
        if type(value) in (int, float):
            self._data["current"] = self._check_and_start_timer(self._enforce_boundaries(value))

    @current.deleter
    def current(self):
        """reset back to base"""
        self._data["current"] = self.base

    @property
    def value(self):
        "The value of the Trait. (current + mod) * mult"
        return self._enforce_boundaries((self.current + self.mod) * self.mult)

    @property
    def ratetarget(self):
        return self._data["ratetarget"]

    @ratetarget.setter
    def ratetarget(self, value):
        self._data["ratetarget"] = self._enforce_boundaries(value)
        self._check_and_start_timer(self.value)

    def percent(self, formatting="{:3.1f}%"):
        """
        Return the current value as a percentage.

        Args:
            formatting (str, optional): Should contain a
               format-tag which will receive the value. If
               this is set to None, the raw float will be
               returned.
        Returns:
            float or str: Depending of if a `formatting` string
                is supplied or not.
        """
        return percent(self.value, self.min, self.max, formatting=formatting)

    def reset(self):
        """Resets `current` property equal to `base` value."""
        del self.current

    def desc(self):
        """
        Retrieve descriptions of the current value, if available.

        This must be a mapping {upper_bound_inclusive: text},
        ordered from small to big. Any value above the highest
        upper bound will be included as being in the highest bound.
        rely on Python3.7+ dicts retaining ordering to let this
        describe the interval.

        Returns:
            str: The description describing the `value` value.
                If not found, returns the empty string.
        """
        descs = self._data["descs"]
        if descs is None:
            return ""
        value = self.value
        # we rely on Python3.7+ dicts retaining ordering
        highest = ""
        for bound, txt in descs.items():
            highest = txt
            if value <= bound:
                return txt
        # if we get here we are above the highest bound so
        # we return the latest bound specified.
        return highest


class GaugeTrait(CounterTrait):
    """
    Gauge Trait.

    This emulates a gauge-meter that empties from a (base+mod) * mult value.

    min/0                                       max=(base+mod)*mult
     |-----------------------X---------------------------|
                           value
                          = current

    - min defaults to 0
    - max value is always (base + mod) * mult
    - .max is an alias of .base
    - value = current and varies from min to max.
    - descs is a mapping {upper_bound_inclusive: desc}. These
        are checked with .desc() and can be retrieve a text
        description for a given current value.

        For example, this could be used to describe health
        values between 0 and 100:
            {0: "Dead"
             10: "Badly hurt",
             30: "Bleeding",
             50: "Hurting",
             90: "Healthy"}

    """

    trait_type = "gauge"

    # same as Counter, here for easy reference
    # current starts out equal to base
    default_keys = {
        "base": 0,
        "mod": 0,
        "mult": 1.0,
        "min": 0,
        "descs": None,
        "rate": 0,
        "ratetarget": None,
    }

    def _update_current(self, current):
        """Update current value by scaling with rate and time passed."""
        rate = self.rate
        if rate != 0 and self._data["last_update"] is not None:
            now = time()
            tdiff = now - self._data["last_update"]
            current += rate * tdiff
            value = current

            # we don't worry about .mod for gauges

            if self._passed_ratetarget(value):
                current = self._data["ratetarget"]
                self._stop_timer()
            elif not self._within_boundaries(value):
                current = self._enforce_boundaries(value)
                self._stop_timer()
            else:
                self._data["last_update"] = now

            self._data["current"] = current

        if self.base is not None and isinstance(self.base, int):
            return round(current)

        return current

    def _enforce_boundaries(self, value):
        """Ensures that incoming value falls within trait's range."""
        if self.min is not None and value <= self.min:
            return self.min
        return min((self.mod + self.base) * self.mult, value)

    def __str__(self):
        status = "{value:4} / {base:4}".format(value=self.value, base=self.base)
        return "{name:12} {status} ({mod:+3}) (* {mult:.2f})".format(
            name=self.name, status=status, mod=self.mod, mult=self.mult
        )

    @property
    def base(self):
        return self._data["base"]

    @base.setter
    def base(self, value):
        """Limit so base+mod can never go below min."""
        if type(value) in (int, float):
            if value + self.mod < self.min:
                value = self.min - self.mod
            self._data["base"] = value

    @property
    def mod(self):
        return self._data["mod"]

    @mod.setter
    def mod(self, value):
        """Limit so base+mod can never go below min."""
        if type(value) in (int, float):
            if value + self.base < self.min:
                value = self.min - self.base
            self._data["mod"] = value

    @property
    def mult(self):
        return self._data["mult"]

    @mult.setter
    def mult(self, amount):
        if type(amount) in (int, float):
            self._data["mult"] = amount

    @mult.deleter
    def mult(self):
        self._data["mult"] = 1.0

    @property
    def min(self):
        val = self._data["min"]
        return self.default_keys["min"] if val is None else val

    @min.setter
    def min(self, value):
        """Limit so min can never be greater than (base+mod)*mult."""
        if value is None:
            self._data["min"] = self.default_keys["min"]
        elif type(value) in (int, float):
            self._data["min"] = min(value, (self.base + self.mod) * self.mult)

    @property
    def max(self):
        "The max is always (base + mod) * mult."
        return (self.base + self.mod) * self.mult

    @max.setter
    def max(self, value):
        raise TraitException(
            "The .max property is not settable on GaugeTraits. Set .mod and .base instead."
        )

    @max.deleter
    def max(self):
        raise TraitException(
            "The .max property cannot be reset on GaugeTraits. Reset .mod and .base instead."
        )

    @property
    def current(self):
        """The `current` value of the gauge."""
        return self._update_current(
            self._enforce_boundaries(self._data.get("current", (self.base + self.mod) * self.mult))
        )

    @current.setter
    def current(self, value):
        if type(value) in (int, float):
            self._data["current"] = self._check_and_start_timer(self._enforce_boundaries(value))

    @current.deleter
    def current(self):
        "Resets current back to 'full'"
        self._data["current"] = (self.base + self.mod) * self.mult

    @property
    def value(self):
        "The value of the trait"
        return self.current

    def percent(self, formatting="{:3.1f}%"):
        """
        Return the current value as a percentage.

        Args:
            formatting (str, optional): Should contain a
               format-tag which will receive the value. If
               this is set to None, the raw float will be
               returned.
        Returns:
            float or str: Depending of if a `formatting` string
                is supplied or not.
        """
        return percent(self.current, self.min, self.max, formatting=formatting)

    def reset(self):
        """
        Fills the gauge to its maximum allowed by base + mod
        """
        del self.current



"""
Email-based login system

Evennia contrib - Griatch 2012


This is a variant of the login system that requires an email-address
instead of a username to login.

This used to be the default Evennia login before replacing it with a
more standard username + password system (having to supply an email
for some reason caused a lot of confusion when people wanted to expand
on it. The email is not strictly needed internally, nor is any
confirmation email sent out anyway).


Installation is simple:

To your settings file, add/edit settings as follows:

```python
CMDSET_UNLOGGEDIN = "contrib.base_systems.email_login.email_login.UnloggedinCmdSet"
CONNECTION_SCREEN_MODULE = "contrib.base_systems.email_login.connection_screens"

```

That's it. Reload the server and try to log in to see it.

The initial login "graphic" will still not mention email addresses
after this change. The login splashscreen is taken from strings in
the module given by settings.CONNECTION_SCREEN_MODULE.

"""

from django.conf import settings

from evennia.accounts.models import AccountDB
from evennia.commands.cmdhandler import CMD_LOGINSTART
from evennia.commands.cmdset import CmdSet
from evennia.commands.default.muxcommand import MuxCommand
from evennia.server.models import ServerConfig
from evennia.utils import ansi, class_from_module, utils

# limit symbol import for API
__all__ = (
    "CmdUnconnectedConnect",
    "CmdUnconnectedCreate",
    "CmdUnconnectedQuit",
    "CmdUnconnectedLook",
    "CmdUnconnectedHelp",
)

CONNECTION_SCREEN_MODULE = settings.CONNECTION_SCREEN_MODULE
CONNECTION_SCREEN = ""
try:
    CONNECTION_SCREEN = ansi.parse_ansi(utils.random_string_from_module(CONNECTION_SCREEN_MODULE))
except Exception:
    # malformed connection screen or no screen given
    pass
if not CONNECTION_SCREEN:
    CONNECTION_SCREEN = (
        "\nEvennia: Error in CONNECTION_SCREEN MODULE"
        " (randomly picked connection screen variable is not a string). \nEnter 'help' for aid."
    )


class CmdUnconnectedConnect(MuxCommand):
    """
    Connect to the game.

    Usage (at login screen):
        connect <email> <password>

    Use the create command to first create an account before logging in.
    """

    key = "connect"
    aliases = ["conn", "con", "co"]
    locks = "cmd:all()"  # not really needed

    def func(self):
        """
        Uses the Django admin api. Note that unlogged-in commands
        have a unique position in that their `func()` receives
        a session object instead of a `source_object` like all
        other types of logged-in commands (this is because
        there is no object yet before the account has logged in)
        """

        session = self.caller
        arglist = self.arglist

        if not arglist or len(arglist) < 2:
            session.msg("\n\r Usage (without <>): connect <email> <password>")
            return
        email = arglist[0]
        password = arglist[1]

        # Match an email address to an account.
        account = AccountDB.objects.get_account_from_email(email)
        # No accountname match
        if not account:
            string = "The email '%s' does not match any accounts." % email
            string += "\n\r\n\rIf you are new you should first create a new account "
            string += "using the 'create' command."
            session.msg(string)
            return
        # We have at least one result, so we can check the password.
        if not account[0].check_password(password):
            session.msg("Incorrect password.")
            return

        # Check IP and/or name bans
        bans = ServerConfig.objects.conf("server_bans")
        if bans and (
            any(tup[0] == account.name for tup in bans)
            or any(tup[2].match(session.address[0]) for tup in bans if tup[2])
        ):
            # this is a banned IP or name!
            string = "|rYou have been banned and cannot continue from here."
            string += "\nIf you feel this ban is in error, please email an admin.|x"
            session.msg(string)
            session.execute_cmd("quit")
            return

        # actually do the login. This will call all hooks.
        session.sessionhandler.login(session, account)


class CmdUnconnectedCreate(MuxCommand):
    """
    Create a new account.

    Usage (at login screen):
        create \"accountname\" <email> <password>

    This creates a new account account.

    """

    key = "create"
    aliases = ["cre", "cr"]
    locks = "cmd:all()"

    def at_pre_cmd(self):
        """Verify that account creation is enabled."""
        if not settings.NEW_ACCOUNT_REGISTRATION_ENABLED:
            # truthy return cancels the command
            self.msg("Registration is currently disabled.")
            return True

        return super().at_pre_cmd()

    def parse(self):
        """
        The parser must handle the multiple-word account
        name enclosed in quotes:
            connect "Long name with many words" my@myserv.com mypassw
        """
        super().parse()

        self.accountinfo = []
        if len(self.arglist) < 3:
            return
        if len(self.arglist) > 3:
            # this means we have a multi_word accountname. pop from the back.
            password = self.arglist.pop()
            email = self.arglist.pop()
            # what remains is the username.
            username = " ".join(self.arglist)
        else:
            username, email, password = self.arglist

        username = username.replace('"', "")  # remove "
        username = username.replace("'", "")
        self.accountinfo = (username, email, password)

    def func(self):
        """Do checks and create account"""

        Account = class_from_module(settings.BASE_ACCOUNT_TYPECLASS)
        address = self.session.address

        session = self.caller
        try:
            username, email, password = self.accountinfo
        except ValueError:
            string = '\n\r Usage (without <>): create "<accountname>" <email> <password>'
            session.msg(string)
            return
        if not email or not password:
            session.msg("\n\r You have to supply an e-mail address followed by a password.")
            return
        if not utils.validate_email_address(email):
            # check so the email at least looks ok.
            session.msg("'%s' is not a valid e-mail address." % email)
            return

        # pre-normalize username so the user know what they get
        non_normalized_username = username
        username = Account.normalize_username(username)
        if non_normalized_username != username:
            session.msg(
                "Note: your username was normalized to strip spaces and remove characters "
                "that could be visually confusing."
            )

        # have the user verify their new account was what they intended
        answer = yield (
            f"You want to create an account '{username}' with email '{email}' and password "
            f"'{password}'.\nIs this what you intended? [Y]/N?"
        )
        if answer.lower() in ("n", "no"):
            session.msg("Aborted. If your user name contains spaces, surround it by quotes.")
            return

        # everything's ok. Create the new player account.
        account, errors = Account.create(
            username=username, email=email, password=password, ip=address, session=session
        )
        if account:
            # tell the caller everything went well.
            string = "A new account '%s' was created. Welcome!"
            if " " in username:
                string += (
                    "\n\nYou can now log in with the command 'connect \"%s\" <your password>'."
                )
            else:
                string += "\n\nYou can now log with the command 'connect %s <your password>'."
            session.msg(string % (username, username))
        else:
            session.msg("|R%s|n" % "\n".join(errors))


class CmdUnconnectedQuit(MuxCommand):
    """
    We maintain a different version of the `quit` command
    here for unconnected accounts for the sake of simplicity. The logged in
    version is a bit more complicated.
    """

    key = "quit"
    aliases = ["q", "qu"]
    locks = "cmd:all()"

    def func(self):
        """Simply close the connection."""
        session = self.caller
        session.sessionhandler.disconnect(session, "Good bye! Disconnecting.")


class CmdUnconnectedLook(MuxCommand):
    """
    This is an unconnected version of the `look` command for simplicity.

    This is called by the server and kicks everything in gear.
    All it does is display the connect screen.
    """

    key = CMD_LOGINSTART
    aliases = ["look", "l"]
    locks = "cmd:all()"

    def func(self):
        """Show the connect screen."""
        self.caller.msg(CONNECTION_SCREEN)


class CmdUnconnectedHelp(MuxCommand):
    """
    This is an unconnected version of the help command,
    for simplicity. It shows a pane of info.
    """

    key = "help"
    aliases = ["h", "?"]
    locks = "cmd:all()"

    def func(self):
        """Shows help"""

        string = """
You are not yet logged into the game. Commands available at this point:
  |wcreate, connect, look, help, quit|n

To login to the system, you need to do one of the following:

|w1)|n If you have no previous account, you need to use the 'create'
   command like this:

     |wcreate "Anna the Barbarian" anna@myemail.com c67jHL8p|n

   It's always a good idea (not only here, but everywhere on the net)
   to not use a regular word for your password. Make it longer than
   3 characters (ideally 6 or more) and mix numbers and capitalization
   into it.

|w2)|n If you have an account already, either because you just created
   one in |w1)|n above or you are returning, use the 'connect' command:

     |wconnect anna@myemail.com c67jHL8p|n

   This should log you in. Run |whelp|n again once you're logged in
   to get more aid. Hope you enjoy your stay!

You can use the |wlook|n command if you want to see the connect screen again.
"""
        self.caller.msg(string)


# command set for the mux-like login


class UnloggedinCmdSet(CmdSet):
    """
    Sets up the unlogged cmdset.
    """

    key = "Unloggedin"
    priority = 0

    def at_cmdset_creation(self):
        """Populate the cmdset"""
        self.add(CmdUnconnectedConnect())
        self.add(CmdUnconnectedCreate())
        self.add(CmdUnconnectedQuit())
        self.add(CmdUnconnectedLook())
        self.add(CmdUnconnectedHelp())



"""
Random Name Generator

Contribution by InspectorCaracal (2022)

A module for generating random names, both real-world and fantasy. Real-world
names can be generated either as first (personal) names, family (last) names, or
full names (first, optional middles, and last). The name data is from [Behind the Name](https://www.behindthename.com/)
and used under the [CC BY-SA 4.0 license](https://creativecommons.org/licenses/by-sa/4.0/).

Fantasy names are generated from basic phonetic rules, using CVC syllable syntax.

Both real-world and fantasy name generation can be extended to include additional
information via your game's `settings.py`


Available Methods:

  first_name   - Selects a random a first (personal) name from the name lists.
  last_name    - Selects a random last (family) name from the name lists.
  full_name    - Generates a randomized full name, optionally including middle names, by selecting first/last names from the name lists.
  fantasy_name - Generates a completely new made-up name based on phonetic rules.

Method examples:

>>> namegen.first_name(num=5)
['Genesis', 'Tali', 'Budur', 'Dominykas', 'Kamau']

>>> namegen.full_name(parts=3, surname_first=True)
'Ó Muircheartach Torunn Dyson'
>>> namegen.full_name(gender='f')
'Wikolia Ó Deasmhumhnaigh'

>>> namegen.fantasy_name(num=3, style="fluid")
['Aewalisash', 'Ayi', 'Iaa']


Available Settings (define these in your `settings.py`)

  NAMEGEN_FIRST_NAMES   - Option to add a new list of first (personal) names.
  NAMEGEN_LAST_NAMES    - Option to add a new list of last (family) names.
  NAMEGEN_REPLACE_LISTS - Set to True if you want to use ONLY your name lists and not the ones that come with the contrib.
  NAMEGEN_FANTASY_RULES - Option to add new fantasy-name style rules.
      Must be a dictionary that includes "syllable", "consonants", "vowels", and "length" - see the example.
      "start" and "end" keys are optional.

Settings examples:

NAMEGEN_FIRST_NAMES = [
        ("Evennia", 'mf'),
        ("Green Tea", 'f'),
    ]

NAMEGEN_LAST_NAMES = [ "Beeblebrox", "Son of Odin" ]

NAMEGEN_FANTASY_RULES = {
    "example_style": {
            "syllable": "(C)VC",
            "consonants": [ 'z','z','ph','sh','r','n' ],
            "start": ['m'],
            "end": ['x','n'],
            "vowels": [ "e","e","e","a","i","i","u","o", ],
            "length": (2,4),
        }
    }

"""

import random
import re
from os import path

from django.conf import settings

from evennia.utils.utils import is_iter

# Load name data from Behind the Name lists
dirpath = path.dirname(path.abspath(__file__))
_FIRSTNAME_LIST = []
with open(path.join(dirpath, "btn_givennames.txt"), "r", encoding="utf-8") as file:
    _FIRSTNAME_LIST = [
        line.strip().rsplit(" ") for line in file if line and not line.startswith("#")
    ]

_SURNAME_LIST = []
with open(path.join(dirpath, "btn_surnames.txt"), "r", encoding="utf-8") as file:
    _SURNAME_LIST = [line.strip() for line in file if line and not line.startswith("#")]

_REQUIRED_KEYS = {"syllable", "consonants", "vowels", "length"}
# Define phoneme structure for built-in fantasy name generators.
_FANTASY_NAME_STRUCTURES = {
    "harsh": {
        "syllable": "CV(C)",
        "consonants": [
            "k",
            "k",
            "k",
            "z",
            "zh",
            "g",
            "v",
            "t",
            "th",
            "w",
            "n",
            "d",
            "d",
        ],
        "start": [
            "dh",
            "kh",
            "kh",
            "kh",
            "vh",
        ],
        "end": [
            "n",
            "x",
        ],
        "vowels": [
            "o",
            "o",
            "o",
            "a",
            "y",
            "u",
            "u",
            "u",
            "ä",
            "ö",
            "e",
            "i",
            "i",
        ],
        "length": (1, 3),
    },
    "fluid": {
        "syllable": "V(C)",
        "consonants": [
            "r",
            "r",
            "l",
            "l",
            "l",
            "l",
            "s",
            "s",
            "s",
            "sh",
            "m",
            "n",
            "n",
            "f",
            "v",
            "w",
            "th",
        ],
        "start": [],
        "end": [],
        "vowels": [
            "a",
            "a",
            "a",
            "a",
            "a",
            "e",
            "i",
            "i",
            "i",
            "y",
            "u",
            "o",
        ],
        "length": (3, 5),
    },
    "alien": {
        "syllable": "C(C(V))(')(C)",
        "consonants": ["q", "q", "x", "z", "v", "w", "k", "h", "b"],
        "start": [
            "x",
        ],
        "end": [],
        "vowels": ["y", "w", "o", "y"],
        "length": (1, 5),
    },
}

_RE_DOUBLES = re.compile(r"(\w)\1{2,}")

# Load in optional settings

custom_first_names = (
    settings.NAMEGEN_FIRST_NAMES if hasattr(settings, "NAMEGEN_FIRST_NAMES") else []
)
custom_last_names = settings.NAMEGEN_LAST_NAMES if hasattr(settings, "NAMEGEN_LAST_NAMES") else []

if hasattr(settings, "NAMEGEN_FANTASY_RULES"):
    _FANTASY_NAME_STRUCTURES |= settings.NAMEGEN_FANTASY_RULES

if hasattr(settings, "NAMEGEN_REPLACE_LISTS") and settings.NAMEGEN_REPLACE_LISTS:
    _FIRSTNAME_LIST = custom_first_names or _FIRSTNAME_LIST
    _SURNAME_LIST = custom_last_names or _SURNAME_LIST

else:
    _FIRSTNAME_LIST += custom_first_names
    _SURNAME_LIST += custom_last_names


def fantasy_name(num=1, style="harsh", return_list=False):
    """
    Generate made-up names in one of a number of "styles".

    Keyword args:
        num (int)      - How many names to return.
        style (string) - The "style" of name. This references an existing algorithm.
        return_list (bool) - Whether to always return a list. `False` by default,
                which returns a string if there is only one value and a list if more.
    """

    def _validate(style_name):
        if style_name not in _FANTASY_NAME_STRUCTURES:
            raise ValueError(
                f"Invalid style name: '{style_name}'. Available style names: {' '.join(_FANTASY_NAME_STRUCTURES.keys())}"
            )
        style_dict = _FANTASY_NAME_STRUCTURES[style_name]

        if type(style_dict) is not dict:
            raise ValueError(f"Style {style_name} must be a dictionary.")

        keys = set(style_dict.keys())
        missing_keys = _REQUIRED_KEYS - keys
        if len(missing_keys):
            raise KeyError(
                f"Style dictionary {style_name} is missing required keys: {' '.join(missing_keys)}"
            )

        if not (type(style_dict["consonants"]) is list and type(style_dict["vowels"]) is list):
            raise TypeError(f"'consonants' and 'vowels' for style {style_name} must be lists.")

        if not (is_iter(style_dict["length"]) and len(style_dict["length"]) == 2):
            raise ValueError(
                f"'length' key for {style_name} must have a minimum and maximum number of syllables."
            )

        return style_dict

    # validate num first
    num = int(num)
    if num < 1:
        raise ValueError("Number of names to generate must be positive.")

    style_dict = _validate(style)

    syllable = []
    weight = 8
    # parse out the syllable structure with weights
    for key in style_dict["syllable"]:
        # parentheses mean optional - allow nested parens
        if key == "(":
            weight = weight / 2
        elif key == ")":
            weight = weight * 2
        else:
            if key == "C":
                sound_type = "consonants"
            elif key == "V":
                sound_type = "vowels"
            else:
                sound_type = key
            # append the sound type and weight
            syllable.append((sound_type, int(weight)))

    name_list = []

    # time to generate a name!
    for n in range(num):
        # build a list of syllables
        length = random.randint(*style_dict["length"])
        name = ""
        for i in range(length):
            # build the syllable itself
            syll = ""
            for sound, weight in syllable:
                # random chance to skip this key; lower weights mean less likely
                if random.randint(0, 8) > weight:
                    continue

                if sound not in style_dict:
                    # extra character, like apostrophes
                    syll += sound
                    continue

                # get a random sound from the sound list
                choices = list(style_dict[sound])

                if sound == "consonants":
                    # if it's a starting consonant, add starting-sounds to the options
                    if not len(syll):
                        choices += style_dict.get("start", [])
                    # if it's an ending consonant, add ending-sounds to the options
                    elif i + 1 == length:
                        choices += style_dict.get("end", [])

                syll += random.choice(choices)

            name += syll

        # condense repeating letters down to a maximum of 2
        name = _RE_DOUBLES.sub(lambda m: m.group(1) * 2, name)
        # capitalize the first letter
        name = name[0].upper() + name[1:] if len(name) > 1 else name.upper()
        name_list.append(name)

    if len(name_list) == 1 and not return_list:
        return name_list[0]
    return name_list


def first_name(
    num=1,
    gender=None,
    return_list=False,
):
    """
    Generate first names, also known as personal names.

    Keyword args:
        num (int)    - How many names to return.
        gender (str) - Restrict names by gender association. `None` by default, which selects from
                all possible names. Set to "m" for masculine, "f" for feminine, "mf" for androgynous
        return_list (bool) - Whether to always return a list. `False` by default,
                which returns a string if there is only one value and a list if more.
    """
    # validate num first
    num = int(num)
    if num < 1:
        raise ValueError("Number of names to generate must be positive.")

    if gender:
        # filter the options by gender
        name_options = [
            name_data[0]
            for name_data in _FIRSTNAME_LIST
            if all([gender_key in gender for gender_key in name_data[1]])
        ]
        if not len(name_options):
            raise ValueError(f"Invalid gender '{gender}'.")
    else:
        name_options = [name_data[0] for name_data in _FIRSTNAME_LIST]

    # take a random selection of `num` names, without repeats
    results = random.sample(name_options, num)

    if len(results) == 1 and not return_list:
        # return single value as a string
        return results[0]

    return results


def last_name(num=1, return_list=False):
    """
    Generate family names, also known as surnames or last names.

    Keyword args:
        num (int)    - How many names to return.
        return_list (bool) - Whether to always return a list. `False` by default,
                which returns a string if there is only one value and a list if more.
    """
    # validate num first
    num = int(num)
    if num < 1:
        raise ValueError("Number of names to generate must be positive.")

    # take a random selection of `num` names, without repeats
    results = random.sample(_SURNAME_LIST, num)

    if len(results) == 1 and not return_list:
        # return single value as a string
        return results[0]

    return results


def full_name(num=1, parts=2, gender=None, return_list=False, surname_first=False):
    """
    Generate complete names with a personal name, family name, and optionally middle names.

    Keyword args:
        num (int)    - How many names to return.
        parts (int)  - How many parts the name should have. By default two: first and last.
        gender (str) - Restrict names by gender association. `None` by default, which selects from
                all possible names. Set to "m" for masculine, "f" for feminine, "mf" for androgynous
        return_list (bool) - Whether to always return a list. `False` by default,
                which returns a string if there is only one value and a list if more.
        surname_first (bool) - Default `False`. Set to `True` if you want the family name to be
                placed at the beginning of the name instead of the end.
    """
    # validate num first
    num = int(num)
    if num < 1:
        raise ValueError("Number of names to generate must be positive.")
    # validate parts next
    parts = int(parts)
    if parts < 2:
        raise ValueError("Number of name parts to generate must be at least 2.")

    name_lists = []

    middle = parts - 2
    if middle:
        # calculate "middle" names.
        # we want them to be an intelligent mix of personal names and family names
        # first, split the total number of middle-name parts into "personal" and "family" at a random point
        total_mids = middle * num
        personals = random.randint(1, total_mids)
        familys = total_mids - personals
        # then get the names for each
        personal_mids = first_name(num=personals, gender=gender, return_list=True)
        family_mids = last_name(num=familys, return_list=True) if familys else []
        # splice them together according to surname_first....
        middle_names = family_mids + personal_mids if surname_first else personal_mids + family_mids
        # ...and then split into `num`-length lists to be used for the final names
        name_lists = [middle_names[num * i : num * (i + 1)] for i in range(0, middle)]

    # get personal and family names
    personal_names = first_name(num=num, gender=gender, return_list=True)
    last_names = last_name(num=num, return_list=True)

    # attach personal/family names to the list of name lists, according to surname_first
    if surname_first:
        name_lists = [last_names] + name_lists + [personal_names]
    else:
        name_lists = [personal_names] + name_lists + [last_names]

    # lastly, zip them all up and join them together
    names = list(zip(*name_lists))
    names = [" ".join(name) for name in names]

    if len(names) == 1 and not return_list:
        # return single value as a string
        return names[0]

    return names



"""
TutorialWorld - basic objects - Griatch 2011

This module holds all "dead" object definitions for
the tutorial world. Object-commands and -cmdsets
are also defined here, together with the object.

Objects:

TutorialObject

TutorialReadable
TutorialClimbable
Obelisk
LightSource
CrumblingWall
TutorialWeapon
TutorialWeaponRack

"""

import random

from evennia import CmdSet, Command, DefaultExit, DefaultObject
from evennia.prototypes.spawner import spawn
from evennia.utils import dedent, delay, search

# -------------------------------------------------------------
#
# TutorialObject
#
# The TutorialObject is the base class for all items
# in the tutorial. They have an attribute "tutorial_info"
# on them that the global tutorial command can use to extract
# interesting behind-the scenes information about the object.
#
# TutorialObjects may also be "reset". What the reset means
# is up to the object. It can be the resetting of the world
# itself, or the removal of an inventory item from a
# character's inventory when leaving the tutorial, for example.
#
# -------------------------------------------------------------


class TutorialObject(DefaultObject):
    """
    This is the baseclass for all objects in the tutorial.
    """

    def at_object_creation(self):
        """Called when the object is first created."""
        super().at_object_creation()
        self.db.tutorial_info = "No tutorial info is available for this object."

    def reset(self):
        """Resets the object, whatever that may mean."""
        self.location = self.home


# -------------------------------------------------------------
#
# Readable - an object that can be "read"
#
# -------------------------------------------------------------

#
# Read command
#


class CmdRead(Command):
    """
    Usage:
      read [obj]

    Read some text of a readable object.
    """

    key = "read"
    locks = "cmd:all()"
    help_category = "TutorialWorld"

    def func(self):
        """
        Implements the read command. This simply looks for an
        Attribute "readable_text" on the object and displays that.
        """

        if self.args:
            obj = self.caller.search(self.args.strip())
        else:
            obj = self.obj
        if not obj:
            return
        # we want an attribute read_text to be defined.
        readtext = obj.db.readable_text
        if readtext:
            string = "You read |C%s|n:\n  %s" % (obj.key, readtext)
        else:
            string = "There is nothing to read on %s." % obj.key
        self.caller.msg(string)


class CmdSetReadable(CmdSet):
    """
    A CmdSet for readables.
    """

    def at_cmdset_creation(self):
        """
        Called when the cmdset is created.
        """
        self.add(CmdRead())


class TutorialReadable(TutorialObject):
    """
    This simple object defines some attributes and
    """

    def at_object_creation(self):
        """
        Called when object is created. We make sure to set the needed
        Attribute and add the readable cmdset.
        """
        super().at_object_creation()
        self.db.tutorial_info = (
            "This is an object with a 'read' command defined in a command set on itself."
        )
        self.db.readable_text = "There is no text written on %s." % self.key
        # define a command on the object.
        self.cmdset.add_default(CmdSetReadable, persistent=True)


# -------------------------------------------------------------
#
# Climbable object
#
# The climbable object works so that once climbed, it sets
# a flag on the climber to show that it was climbed. A simple
# command 'climb' handles the actual climbing. The memory
# of what was last climbed is used in a simple puzzle in the
# tutorial.
#
# -------------------------------------------------------------


class CmdClimb(Command):
    """
    Climb an object

    Usage:
      climb <object>

    This allows you to climb.
    """

    key = "climb"
    locks = "cmd:all()"
    help_category = "TutorialWorld"

    def func(self):
        """Implements function"""

        if not self.args:
            self.caller.msg("What do you want to climb?")
            return
        obj = self.caller.search(self.args.strip())
        if not obj:
            return
        if obj != self.obj:
            self.caller.msg("Try as you might, you cannot climb that.")
            return
        ostring = self.obj.db.climb_text
        if not ostring:
            ostring = "You climb %s. Having looked around, you climb down again." % self.obj.name
        self.caller.msg(ostring)
        # set a tag on the caller to remember that we climbed.
        self.caller.tags.add("tutorial_climbed_tree", category="tutorial_world")


class CmdSetClimbable(CmdSet):
    """Climbing cmdset"""

    def at_cmdset_creation(self):
        """populate set"""
        self.add(CmdClimb())


class TutorialClimbable(TutorialObject):
    """
    A climbable object. All that is special about it is that it has
    the "climb" command available on it.
    """

    def at_object_creation(self):
        """Called at initial creation only"""
        self.cmdset.add_default(CmdSetClimbable, persistent=True)


# -------------------------------------------------------------
#
# Obelisk - a unique item
#
# The Obelisk is an object with a modified return_appearance method
# that causes it to look slightly different every time one looks at it.
# Since what you actually see is a part of a game puzzle, the act of
# looking also stores a key attribute on the looking object (different
# depending on which text you saw) for later reference.
#
# -------------------------------------------------------------


class Obelisk(TutorialObject):
    """
    This object changes its description randomly, and which is shown
    determines which order "clue id" is stored on the Character for
    future puzzles.

    Important Attribute:
       puzzle_descs (list): list of descriptions. One of these is
        picked randomly when this object is looked at and its index
        in the list is used as a key for to solve the puzzle.

    """

    def at_object_creation(self):
        """Called when object is created."""
        super().at_object_creation()
        self.db.tutorial_info = (
            "This object changes its desc randomly, and makes sure to remember which one you saw."
        )
        self.db.puzzle_descs = ["You see a normal stone slab"]
        # make sure this can never be picked up
        self.locks.add("get:false()")

    def return_appearance(self, caller):
        """
        This hook is called by the look command to get the description
        of the object. We overload it with our own version.
        """
        # randomly get the index for one of the descriptions
        descs = self.db.puzzle_descs
        clueindex = random.randint(0, len(descs) - 1)
        # set this description, with the random extra
        string = (
            "The surface of the obelisk seem to waver, shift and writhe under your gaze, with "
            "different scenes and structures appearing whenever you look at it. "
        )
        self.db.desc = string + descs[clueindex]
        # remember that this was the clue we got. The Puzzle room will
        # look for this later to determine if you should be teleported
        # or not.
        caller.db.puzzle_clue = clueindex
        # call the parent function as normal (this will use
        # the new desc Attribute we just set)
        return super().return_appearance(caller)


# -------------------------------------------------------------
#
# LightSource
#
# This object emits light. Once it has been turned on it
# cannot be turned off. When it burns out it will delete
# itself.
#
# This could be implemented using a single-repeat Script or by
# registering with the TickerHandler. We do it simpler by
# using the delay() utility function. This is very simple
# to use but does not survive a server @reload. Because of
# where the light matters (in the Dark Room where you can
# find new light sources easily), this is okay here.
#
# -------------------------------------------------------------


class CmdLight(Command):
    """
    Creates light where there was none. Something to burn.
    """

    key = "on"
    aliases = ["light", "burn"]
    # only allow this command if command.obj is carried by caller.
    locks = "cmd:holds()"
    help_category = "TutorialWorld"

    def func(self):
        """
        Implements the light command. Since this command is designed
        to sit on a "lightable" object, we operate only on self.obj.
        """

        if self.obj.light():
            self.caller.msg("You light %s." % self.obj.key)
            self.caller.location.msg_contents(
                "%s lights %s!" % (self.caller, self.obj.key), exclude=[self.caller]
            )
        else:
            self.caller.msg("%s is already burning." % self.obj.key)


class CmdSetLight(CmdSet):
    """CmdSet for the lightsource commands"""

    key = "lightsource_cmdset"
    # this is higher than the dark cmdset - important!
    priority = 3

    def at_cmdset_creation(self):
        """called at cmdset creation"""
        self.add(CmdLight())


class LightSource(TutorialObject):
    """
    This implements a light source object.

    When burned out, the object will be deleted.
    """

    def at_init(self):
        """
        If this is called with the Attribute is_giving_light already
        set, we know that the timer got killed by a server
        reload/reboot before it had time to finish. So we kill it here
        instead. This is the price we pay for the simplicity of the
        non-persistent delay() method.
        """
        if self.db.is_giving_light:
            self.delete()

    def at_object_creation(self):
        """Called when object is first created."""
        super().at_object_creation()
        self.db.tutorial_info = (
            "This object can be lit to create light. It has a timeout for how long it burns."
        )
        self.db.is_giving_light = False
        self.db.burntime = 60 * 3  # 3 minutes
        # this is the default desc, it can of course be customized
        # when created.
        self.db.desc = "A splinter of wood with remnants of resin on it, enough for burning."
        # add the Light command
        self.cmdset.add_default(CmdSetLight, persistent=True)

    def _burnout(self):
        """
        This is called when this light source burns out. We make no
        use of the return value.
        """
        # delete ourselves from the database
        self.db.is_giving_light = False
        try:
            self.location.location.msg_contents(
                "%s's %s flickers and dies." % (self.location, self.key), exclude=self.location
            )
            self.location.msg("Your %s flickers and dies." % self.key)
            self.location.location.check_light_state()
        except AttributeError:
            try:
                self.location.msg_contents("A %s on the floor flickers and dies." % self.key)
                self.location.location.check_light_state()
            except AttributeError:
                # Mainly happens if we happen to be in a None location
                pass
        self.delete()

    def light(self):
        """
        Light this object - this is called by Light command.
        """
        if self.db.is_giving_light:
            return False
        # burn for 3 minutes before calling _burnout
        self.db.is_giving_light = True
        # if we are in a dark room, trigger its light check
        try:
            self.location.location.check_light_state()
        except AttributeError:
            try:
                # maybe we are directly in the room
                self.location.check_light_state()
            except AttributeError:
                # we are in a None location
                pass
        finally:
            # start the burn timer. When it runs out, self._burnout
            # will be called. We store the deferred so it can be
            # killed in unittesting.
            self.deferred = delay(60 * 3, self._burnout)
        return True


# -------------------------------------------------------------
#
# Crumbling wall - unique exit
#
# This implements a simple puzzle exit that needs to be
# accessed with commands before one can get to traverse it.
#
# The puzzle-part is simply to move roots (that have
# presumably covered the wall) aside until a button for a
# secret door is revealed. The original position of the
# roots blocks the button, so they have to be moved to a certain
# position - when they have, the "press button" command
# is made available and the Exit is made traversable.
#
# -------------------------------------------------------------

# There are four roots - two horizontal and two vertically
# running roots. Each can have three positions: top/middle/bottom
# and left/middle/right respectively. There can be any number of
# roots hanging through the middle position, but only one each
# along the sides. The goal is to make the center position clear.
# (yes, it's really as simple as it sounds, just move the roots
# to each side to "win". This is just a tutorial, remember?)
#
# The ShiftRoot command depends on the root object having an
# Attribute root_pos (a dictionary) to describe the current
# position of the roots.


class CmdShiftRoot(Command):
    """
    Shifts roots around.

    Usage:
      shift blue root left/right
      shift red root left/right
      shift yellow root up/down
      shift green root up/down

    """

    key = "shift"
    aliases = ["shiftroot", "push", "pull", "move"]
    # we only allow to use this command while the
    # room is properly lit, so we lock it to the
    # setting of Attribute "is_lit" on our location.
    locks = "cmd:locattr(is_lit)"
    help_category = "TutorialWorld"

    def parse(self):
        """
        Custom parser; split input by spaces for simplicity.
        """
        self.arglist = self.args.strip().split()

    def func(self):
        """
        Implement the command.
          blue/red - vertical roots
          yellow/green - horizontal roots
        """

        if not self.arglist:
            self.caller.msg("What do you want to move, and in what direction?")
            return

        if "root" in self.arglist:
            # we clean out the use of the word "root"
            self.arglist.remove("root")

        # we accept arguments on the form <color> <direction>

        if not len(self.arglist) > 1:
            self.caller.msg(
                "You must define which colour of root you want to move, and in which direction."
            )
            return

        color = self.arglist[0].lower()
        direction = self.arglist[1].lower()

        # get current root positions dict
        root_pos = self.obj.db.root_pos

        if color not in root_pos:
            self.caller.msg("No such root to move.")
            return

        # first, vertical roots (red/blue) - can be moved left/right
        if color == "red":
            if direction == "left":
                root_pos[color] = max(-1, root_pos[color] - 1)
                self.caller.msg("You shift the reddish root to the left.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["blue"]:
                    root_pos["blue"] += 1
                    self.caller.msg(
                        "The root with blue flowers gets in the way and is pushed to the right."
                    )
            elif direction == "right":
                root_pos[color] = min(1, root_pos[color] + 1)
                self.caller.msg("You shove the reddish root to the right.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["blue"]:
                    root_pos["blue"] -= 1
                    self.caller.msg(
                        "The root with blue flowers gets in the way and is pushed to the left."
                    )
            else:
                self.caller.msg(
                    "The root hangs straight down - you can only move it left or right."
                )
        elif color == "blue":
            if direction == "left":
                root_pos[color] = max(-1, root_pos[color] - 1)
                self.caller.msg("You shift the root with small blue flowers to the left.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["red"]:
                    root_pos["red"] += 1
                    self.caller.msg(
                        "The reddish root is too big to fit as well, so that one falls away to the left."
                    )
            elif direction == "right":
                root_pos[color] = min(1, root_pos[color] + 1)
                self.caller.msg("You shove the root adorned with small blue flowers to the right.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["red"]:
                    root_pos["red"] -= 1
                    self.caller.msg(
                        "The thick reddish root gets in the way and is pushed back to the left."
                    )
            else:
                self.caller.msg(
                    "The root hangs straight down - you can only move it left or right."
                )

        # now the horizontal roots (yellow/green). They can be moved up/down
        elif color == "yellow":
            if direction == "up":
                root_pos[color] = max(-1, root_pos[color] - 1)
                self.caller.msg("You shift the root with small yellow flowers upwards.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["green"]:
                    root_pos["green"] += 1
                    self.caller.msg("The green weedy root falls down.")
            elif direction == "down":
                root_pos[color] = min(1, root_pos[color] + 1)
                self.caller.msg("You shove the root adorned with small yellow flowers downwards.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["green"]:
                    root_pos["green"] -= 1
                    self.caller.msg("The weedy green root is shifted upwards to make room.")
            else:
                self.caller.msg("The root hangs across the wall - you can only move it up or down.")
        elif color == "green":
            if direction == "up":
                root_pos[color] = max(-1, root_pos[color] - 1)
                self.caller.msg("You shift the weedy green root upwards.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["yellow"]:
                    root_pos["yellow"] += 1
                    self.caller.msg("The root with yellow flowers falls down.")
            elif direction == "down":
                root_pos[color] = min(1, root_pos[color] + 1)
                self.caller.msg("You shove the weedy green root downwards.")
                if root_pos[color] != 0 and root_pos[color] == root_pos["yellow"]:
                    root_pos["yellow"] -= 1
                    self.caller.msg(
                        "The root with yellow flowers gets in the way and is pushed upwards."
                    )
            else:
                self.caller.msg("The root hangs across the wall - you can only move it up or down.")

        # we have moved the root. Store new position
        self.obj.db.root_pos = root_pos

        # Check victory condition
        if list(root_pos.values()).count(0) == 0:  # no roots in middle position
            # This will affect the cmd: lock of CmdPressButton
            self.obj.db.button_exposed = True
            self.caller.msg("Holding aside the root you think you notice something behind it ...")


class CmdPressButton(Command):
    """
    Presses a button.
    """

    key = "press"
    aliases = ["press button", "button", "push button"]
    # only accessible if the button was found and there is light. This checks
    # the Attribute button_exposed on the Wall object so that
    # you can only push the button when the puzzle is solved. It also
    # checks the is_lit Attribute on the location.
    locks = "cmd:objattr(button_exposed) and objlocattr(is_lit)"
    help_category = "TutorialWorld"

    def func(self):
        """Implements the command"""

        if self.caller.db.crumbling_wall_found_exit:
            # we already pushed the button
            self.caller.msg(
                "The button folded away when the secret passage opened. You cannot push it again."
            )
            return

        # pushing the button
        string = (
            "You move your fingers over the suspicious depression, then gives it a "
            "decisive push. First nothing happens, then there is a rumble and a hidden "
            "|wpassage|n opens, dust and pebbles rumbling as part of the wall moves aside."
        )
        self.caller.msg(string)
        string = (
            "%s moves their fingers over the suspicious depression, then gives it a "
            "decisive push. First nothing happens, then there is a rumble and a hidden "
            "|wpassage|n opens, dust and pebbles rumbling as part of the wall moves aside."
        )
        self.caller.location.msg_contents(string % self.caller.key, exclude=self.caller)
        if not self.obj.open_wall():
            self.caller.msg("The exit leads nowhere, there's just more stone behind it ...")


class CmdSetCrumblingWall(CmdSet):
    """Group the commands for crumblingWall"""

    key = "crumblingwall_cmdset"
    priority = 2

    def at_cmdset_creation(self):
        """called when object is first created."""
        self.add(CmdShiftRoot())
        self.add(CmdPressButton())


class CrumblingWall(TutorialObject, DefaultExit):
    """
    This is a custom Exit.

    The CrumblingWall can be examined in various ways, but only if a
    lit light source is in the room. The traversal itself is blocked
    by a traverse: lock on the exit that only allows passage if a
    certain attribute is set on the trying account.

    Important attribute
     destination - this property must be set to make this a valid exit
                   whenever the button is pushed (this hides it as an exit
                   until it actually is)
    """

    def at_init(self):
        """
        Called when object is recalled from cache.
        """
        self.reset()

    def at_object_creation(self):
        """called when the object is first created."""
        super().at_object_creation()

        self.aliases.add(["secret passage", "passage", "crack", "opening", "secret"])

        # starting root positions. H1/H2 are the horizontally hanging roots,
        # V1/V2 the vertically hanging ones. Each can have three positions:
        # (-1, 0, 1) where 0 means the middle position. yellow/green are
        # horizontal roots and red/blue vertical, all may have value 0, but n
        # ever any other identical value.
        self.db.root_pos = {"yellow": 0, "green": 0, "red": 0, "blue": 0}

        # flags controlling the puzzle victory conditions
        self.db.button_exposed = False
        self.db.exit_open = False

        # this is not even an Exit until it has a proper destination, and we won't assign
        # that until it is actually open. Until then we store the destination here. This
        # should be given a reasonable value at creation!
        self.db.destination = "#2"

        # we lock this Exit so that one can only execute commands on it
        # if its location is lit and only traverse it once the Attribute
        # exit_open is set to True.
        self.locks.add("cmd:locattr(is_lit);traverse:objattr(exit_open)")
        # set cmdset
        self.cmdset.add(CmdSetCrumblingWall, persistent=True)

    def open_wall(self):
        """
        This method is called by the push button command once the puzzle
        is solved. It opens the wall and sets a timer for it to reset
        itself.
        """
        # this will make it into a proper exit (this returns a list)
        eloc = search.search_object(self.db.destination)
        if not eloc:
            return False
        else:
            self.destination = eloc[0]
        self.db.exit_open = True
        # start a 45 second timer before closing again. We store the deferred so it can be
        # killed in unittesting.
        self.deferred = delay(45, self.reset)
        return True

    def _translate_position(self, root, ipos):
        """Translates the position into words"""
        rootnames = {
            "red": "The |rreddish|n vertical-hanging root ",
            "blue": "The thick vertical root with |bblue|n flowers ",
            "yellow": "The thin horizontal-hanging root with |yyellow|n flowers ",
            "green": "The weedy |ggreen|n horizontal root ",
        }
        vpos = {
            -1: "hangs far to the |wleft|n on the wall.",
            0: "hangs straight down the |wmiddle|n of the wall.",
            1: "hangs far to the |wright|n of the wall.",
        }
        hpos = {
            -1: "covers the |wupper|n part of the wall.",
            0: "passes right over the |wmiddle|n of the wall.",
            1: "nearly touches the floor, near the |wbottom|n of the wall.",
        }

        if root in ("yellow", "green"):
            string = rootnames[root] + hpos[ipos]
        else:
            string = rootnames[root] + vpos[ipos]
        return string

    def return_appearance(self, caller):
        """
        This is called when someone looks at the wall. We need to echo the
        current root positions.
        """
        if self.db.button_exposed:
            # we found the button by moving the roots
            result = [
                "Having moved all the roots aside, you find that the center of the wall, "
                "previously hidden by the vegetation, hid a curious square depression. It was maybe once "
                "concealed and made to look a part of the wall, but with the crumbling of stone around it, "
                "it's now easily identifiable as some sort of button."
            ]
        elif self.db.exit_open:
            # we pressed the button; the exit is open
            result = [
                "With the button pressed, a crack has opened in the root-covered wall, just wide enough "
                "to squeeze through. A cold draft is coming from the hole and you get the feeling the "
                "opening may close again soon."
            ]
        else:
            # puzzle not solved yet.
            result = [
                "The wall is old and covered with roots that here and there have permeated the stone. "
                "The roots (or whatever they are - some of them are covered in small nondescript flowers) "
                "crisscross the wall, making it hard to clearly see its stony surface. Maybe you could "
                "try to |wshift|n or |wmove|n them (like '|wshift red up|n').\n"
            ]
            # display the root positions to help with the puzzle
            for key, pos in self.db.root_pos.items():
                result.append("\n" + self._translate_position(key, pos))
        self.db.desc = "".join(result)

        # call the parent to continue execution (will use the desc we just set)
        return super().return_appearance(caller)

    def at_post_traverse(self, traverser, source_location):
        """
        This is called after we traversed this exit. Cleans up and resets
        the puzzle.
        """
        del traverser.db.crumbling_wall_found_buttothe
        del traverser.db.crumbling_wall_found_exit
        self.reset()

    def at_failed_traverse(self, traverser):
        """This is called if the account fails to pass the Exit."""
        traverser.msg("No matter how you try, you cannot force yourself through %s." % self.key)

    def reset(self):
        """
        Called by tutorial world runner, or whenever someone successfully
        traversed the Exit.
        """
        self.location.msg_contents(
            "The secret door closes abruptly, roots falling back into place."
        )

        # reset the flags and remove the exit destination
        self.db.button_exposed = False
        self.db.exit_open = False
        self.destination = None

        # Reset the roots with some random starting positions for the roots:
        start_pos = [
            {"yellow": 1, "green": 0, "red": 0, "blue": 0},
            {"yellow": 0, "green": 0, "red": 0, "blue": 0},
            {"yellow": 0, "green": 1, "red": -1, "blue": 0},
            {"yellow": 1, "green": 0, "red": 0, "blue": 0},
            {"yellow": 0, "green": 0, "red": 0, "blue": 1},
        ]
        self.db.root_pos = random.choice(start_pos)


# -------------------------------------------------------------
#
# TutorialWeapon - object type
#
# A weapon is necessary in order to fight in the tutorial
# world. A weapon (which here is assumed to be a bladed
# melee weapon for close combat) has three commands,
# stab, slash and defend. Weapons also have a property "magic"
# to determine if they are usable against certain enemies.
#
# Since Characters don't have special skills in the tutorial,
# we let the weapon itself determine how easy/hard it is
# to hit with it, and how much damage it can do.
#
# -------------------------------------------------------------


class CmdAttack(Command):
    """
    Attack the enemy. Commands:

      stab <enemy>
      slash <enemy>
      parry

    stab - (thrust) makes a lot of damage but is harder to hit with.
    slash - is easier to land, but does not make as much damage.
    parry - forgoes your attack but will make you harder to hit on next
            enemy attack.

    """

    # this is an example of implementing many commands as a single
    # command class, using the given command alias to separate between them.

    key = "attack"
    aliases = [
        "hit",
        "kill",
        "fight",
        "thrust",
        "pierce",
        "stab",
        "slash",
        "chop",
        "bash",
        "parry",
        "defend",
    ]
    locks = "cmd:all()"
    help_category = "TutorialWorld"

    def func(self):
        """Implements the stab"""

        cmdstring = self.cmdstring

        if cmdstring in ("attack", "fight"):
            string = "How do you want to fight? Choose one of 'stab', 'slash' or 'defend'."
            self.caller.msg(string)
            return

        # parry mode
        if cmdstring in ("parry", "defend"):
            string = (
                "You raise your weapon in a defensive pose, ready to block the next enemy attack."
            )
            self.caller.msg(string)
            self.caller.db.combat_parry_mode = True
            self.caller.location.msg_contents(
                "%s takes a defensive stance" % self.caller, exclude=[self.caller]
            )
            return

        if not self.args:
            self.caller.msg("Who do you attack?")
            return
        target = self.caller.search(self.args.strip())
        if not target:
            return

        if cmdstring in ("thrust", "pierce", "stab"):
            hit = float(self.obj.db.hit) * 0.7  # modified due to stab
            damage = self.obj.db.damage * 2  # modified due to stab
            string = "You stab with %s. " % self.obj.key
            tstring = "%s stabs at you with %s. " % (self.caller.key, self.obj.key)
            ostring = "%s stabs at %s with %s. " % (self.caller.key, target.key, self.obj.key)
            self.caller.db.combat_parry_mode = False
        elif cmdstring in ("slash", "chop", "bash"):
            hit = float(self.obj.db.hit)  # un modified due to slash
            damage = self.obj.db.damage  # un modified due to slash
            string = "You slash with %s. " % self.obj.key
            tstring = "%s slash at you with %s. " % (self.caller.key, self.obj.key)
            ostring = "%s slash at %s with %s. " % (self.caller.key, target.key, self.obj.key)
            self.caller.db.combat_parry_mode = False
        else:
            self.caller.msg(
                "You fumble with your weapon, unsure of whether to stab, slash or parry ..."
            )
            self.caller.location.msg_contents(
                "%s fumbles with their weapon." % self.caller, exclude=self.caller
            )
            self.caller.db.combat_parry_mode = False
            return

        if target.db.combat_parry_mode:
            # target is defensive; even harder to hit!
            target.msg("|GYou defend, trying to avoid the attack.|n")
            hit *= 0.5

        if random.random() <= hit:
            self.caller.msg(string + "|gIt's a hit!|n")
            target.msg(tstring + "|rIt's a hit!|n")
            self.caller.location.msg_contents(
                ostring + "It's a hit!", exclude=[target, self.caller]
            )

            # call enemy hook
            if hasattr(target, "at_hit"):
                # should return True if target is defeated, False otherwise.
                target.at_hit(self.obj, self.caller, damage)
                return
            elif target.db.health:
                target.db.health -= damage
            else:
                # sorry, impossible to fight this enemy ...
                self.caller.msg("The enemy seems unaffected.")
                return
        else:
            self.caller.msg(string + "|rYou miss.|n")
            target.msg(tstring + "|gThey miss you.|n")
            self.caller.location.msg_contents(ostring + "They miss.", exclude=[target, self.caller])


class CmdSetWeapon(CmdSet):
    """Holds the attack command."""

    def at_cmdset_creation(self):
        """called at first object creation."""
        self.add(CmdAttack())


class TutorialWeapon(TutorialObject):
    """
    This defines a bladed weapon.

    Important attributes (set at creation):
      hit - chance to hit (0-1)
      parry - chance to parry (0-1)
      damage - base damage given (modified by hit success and
               type of attack) (0-10)

    """

    def at_object_creation(self):
        """Called at first creation of the object"""
        super().at_object_creation()
        self.db.hit = 0.4  # hit chance
        self.db.parry = 0.8  # parry chance
        self.db.damage = 1.0
        self.db.magic = False
        self.cmdset.add_default(CmdSetWeapon, persistent=True)

    def reset(self):
        """
        When reset, the weapon is simply deleted, unless it has a place
        to return to.
        """
        if self.location.has_account and self.home == self.location:
            self.location.msg_contents(
                "%s suddenly and magically fades into nothingness, as if it was never there ..."
                % self.key
            )
            self.delete()
        else:
            self.location = self.home


# -------------------------------------------------------------
#
# Weapon rack - spawns weapons
#
# This is a spawner mechanism that creates custom weapons from a
# spawner prototype dictionary. Note that we only create a single typeclass
# (Weapon) yet customize all these different weapons using the spawner.
# The spawner dictionaries could easily sit in separate modules and be
# used to create unique and interesting variations of typeclassed
# objects.
#
# -------------------------------------------------------------

WEAPON_PROTOTYPES = {
    "weapon": {
        "typeclass": "evennia.contrib.tutorials.tutorial_world.objects.TutorialWeapon",
        "key": "Weapon",
        "hit": 0.2,
        "parry": 0.2,
        "damage": 1.0,
        "magic": False,
        "desc": "A generic blade.",
    },
    "knife": {
        "prototype_parent": "weapon",
        "aliases": "sword",
        "key": "Kitchen knife",
        "desc": "A rusty kitchen knife. Better than nothing.",
        "damage": 3,
    },
    "dagger": {
        "prototype_parent": "knife",
        "key": "Rusty dagger",
        "aliases": ["knife", "dagger"],
        "desc": "A double-edged dagger with a nicked edge and a wooden handle.",
        "hit": 0.25,
    },
    "sword": {
        "prototype_parent": "weapon",
        "key": "Rusty sword",
        "aliases": ["sword"],
        "desc": "A rusty shortsword. It has a leather-wrapped handle covered i food grease.",
        "hit": 0.3,
        "damage": 5,
        "parry": 0.5,
    },
    "club": {
        "prototype_parent": "weapon",
        "key": "Club",
        "desc": "A heavy wooden club, little more than a heavy branch.",
        "hit": 0.4,
        "damage": 6,
        "parry": 0.2,
    },
    "axe": {
        "prototype_parent": "weapon",
        "key": "Axe",
        "desc": "A woodcutter's axe with a keen edge.",
        "hit": 0.4,
        "damage": 6,
        "parry": 0.2,
    },
    "ornate longsword": {
        "prototype_parent": "sword",
        "key": "Ornate longsword",
        "desc": "A fine longsword with some swirling patterns on the handle.",
        "hit": 0.5,
        "magic": True,
        "damage": 5,
    },
    "warhammer": {
        "prototype_parent": "club",
        "key": "Silver Warhammer",
        "aliases": ["hammer", "warhammer", "war"],
        "desc": "A heavy war hammer with silver ornaments. This huge weapon causes massive damage - if you can hit.",
        "hit": 0.4,
        "magic": True,
        "damage": 8,
    },
    "rune axe": {
        "prototype_parent": "axe",
        "key": "Runeaxe",
        "aliases": ["axe"],
        "hit": 0.4,
        "magic": True,
        "damage": 6,
    },
    "thruning": {
        "prototype_parent": "ornate longsword",
        "key": "Broadsword named Thruning",
        "desc": "This heavy bladed weapon is marked with the name 'Thruning'. It is very powerful in skilled hands.",
        "hit": 0.6,
        "parry": 0.6,
        "damage": 7,
    },
    "slayer waraxe": {
        "prototype_parent": "rune axe",
        "key": "Slayer waraxe",
        "aliases": ["waraxe", "war", "slayer"],
        "desc": "A huge double-bladed axe marked with the runes for 'Slayer'."
        " It has more runic inscriptions on its head, which you cannot decipher.",
        "hit": 0.7,
        "damage": 8,
    },
    "ghostblade": {
        "prototype_parent": "ornate longsword",
        "key": "The Ghostblade",
        "aliases": ["blade", "ghost"],
        "desc": "This massive sword is large as you are tall, yet seems to weigh almost nothing."
        " It's almost like it's not really there.",
        "hit": 0.9,
        "parry": 0.8,
        "damage": 10,
    },
    "hawkblade": {
        "prototype_parent": "ghostblade",
        "key": "The Hawkblade",
        "aliases": ["hawk", "blade"],
        "desc": "The weapon of a long-dead heroine and a more civilized age,"
        " the hawk-shaped hilt of this blade almost has a life of its own.",
        "hit": 0.85,
        "parry": 0.7,
        "damage": 11,
    },
}


class CmdGetWeapon(Command):
    """
    Usage:
      get weapon

    This will try to obtain a weapon from the container.
    """

    key = "get weapon"
    aliases = "get weapon"
    locks = "cmd:all()"
    help_category = "TutorialWorld"

    def func(self):
        """
        Get a weapon from the container. It will
        itself handle all messages.
        """
        self.obj.produce_weapon(self.caller)


class CmdSetWeaponRack(CmdSet):
    """
    The cmdset for the rack.
    """

    key = "weaponrack_cmdset"

    def at_cmdset_creation(self):
        """Called at first creation of cmdset"""
        self.add(CmdGetWeapon())


class TutorialWeaponRack(TutorialObject):
    """
    This object represents a weapon store. When people use the
    "get weapon" command on this rack, it will produce one
    random weapon from among those registered to exist
    on it. This will also set a property on the character
    to make sure they can't get more than one at a time.

    Attributes to set on this object:
        available_weapons: list of prototype-keys from
            WEAPON_PROTOTYPES, the weapons available in this rack.
        no_more_weapons_msg - error message to return to accounts
            who already got one weapon from the rack and tries to
            grab another one.

    """

    def at_object_creation(self):
        """
        called at creation
        """
        self.cmdset.add_default(CmdSetWeaponRack, persistent=True)
        self.db.rack_id = "weaponrack_1"
        # these are prototype names from the prototype
        # dictionary above.
        self.db.get_weapon_msg = dedent(
            """
            You find |c%s|n. While carrying this weapon, these actions are available:
              |wstab/thrust/pierce <target>|n - poke at the enemy. More damage but harder to hit.
              |wslash/chop/bash <target>|n - swipe at the enemy. Less damage but easier to hit.
              |wdefend/parry|n - protect yourself and make yourself harder to hit.)
            """
        ).strip()

        self.db.no_more_weapons_msg = "you find nothing else of use."
        self.db.available_weapons = ["knife", "dagger", "sword", "club"]

    def produce_weapon(self, caller):
        """
        This will produce a new weapon from the rack,
        assuming the caller hasn't already gotten one. When
        doing so, the caller will get Tagged with the id
        of this rack, to make sure they cannot keep
        pulling weapons from it indefinitely.
        """
        rack_id = self.db.rack_id
        if caller.tags.get(rack_id, category="tutorial_world"):
            caller.msg(self.db.no_more_weapons_msg)
        else:
            prototype = random.choice(self.db.available_weapons)
            # use the spawner to create a new Weapon from the
            # spawner dictionary, tag the caller
            wpn = spawn(WEAPON_PROTOTYPES[prototype], prototype_parents=WEAPON_PROTOTYPES)[0]
            caller.tags.add(rack_id, category="tutorial_world")
            wpn.location = caller
            caller.msg(self.db.get_weapon_msg % wpn.key)



"""
Cooldown contrib module.

Evennia contrib - owllex, 2021

This contrib provides a simple cooldown handler that can be attached to any
typeclassed Object or Account. A cooldown is a lightweight persistent
asynchronous timer that you can query to see if it is ready.

Cooldowns are good for modelling rate-limited actions, like how often a
character can perform a given command.

Cooldowns are completely asynchronous and must be queried to know their
state. They do not fire callbacks, so are not a good fit for use cases
where something needs to happen on a specific schedule (use delay or
a TickerHandler for that instead).

See also the evennia documentation for command cooldowns
(https://github.com/evennia/evennia/wiki/Command-Cooldown) for more information
about the concept.

Installation:

To use, simply add the following property to the typeclass definition of any
object type that you want to support cooldowns. It will expose a new `cooldowns`
property that persists data to the object's attribute storage. You can set this
on your base `Object` typeclass to enable cooldown tracking on every kind of
object, or just put it on your `Character` typeclass.

By default the CooldownHandler will use the `cooldowns` property, but you can
customize this if desired by passing a different value for the db_attribute
parameter.

    from evennia.contrib.game_systems.cooldowns import Cooldownhandler
    from evennia.utils.utils import lazy_property

    @lazy_property
    def cooldowns(self):
        return CooldownHandler(self, db_attribute="cooldowns")

Example:

Assuming you've installed cooldowns on your Character typeclasses, you can use a
cooldown to limit how often you can perform a command. The following code
snippet will limit the use of a Power Attack command to once every 10 seconds
per character.

class PowerAttack(Command):
    def func(self):
        if self.caller.cooldowns.ready("power attack"):
            self.do_power_attack()
            self.caller.cooldowns.add("power attack", 10)
        else:
            self.caller.msg("That's not ready yet!")

"""

import math
import time


class CooldownHandler:
    """
    Handler for cooldowns. This can be attached to any object that supports DB
    attributes (like a Character or Account).

    A cooldown is a timer that is usually used to limit how often some action
    can be performed or some effect can trigger. When a cooldown is first added,
    it counts down from the amount of time provided back to zero, at which point
    it is considered ready again.

    Cooldowns are named with an arbitrary string, and that string is used to
    check on the progression of the cooldown. Each cooldown is tracked
    separately and independently from other cooldowns on that same object. A
    cooldown is unique per-object.

    Cooldowns are saved persistently, so they survive reboots. This module does
    not register or provide callback functionality for when a cooldown becomes
    ready again. Users of cooldowns are expected to query the state of any
    cooldowns they are interested in.

    Methods:
    - ready(name): Checks whether a given cooldown name is ready.
    - time_left(name): Returns how much time is left on a cooldown.
    - add(name, seconds): Sets a given cooldown to last for a certain
        amount of time. Until then, ready() will return False for that
        cooldown name. set() is an alias.
    - extend(name, seconds): Like add(), but adds more time to the given
        cooldown if it already exists. If it doesn't exist yet, calling
        this is equivalent to calling add().
    - reset(cooldown): Resets a given cooldown, causing ready() to return
        True for that cooldown immediately.
    - clear(): Resets all cooldowns.
    """

    __slots__ = ("data", "db_attribute", "obj")

    def __init__(self, obj, db_attribute="cooldowns"):
        if not obj.attributes.has(db_attribute):
            obj.attributes.add(db_attribute, {})

        self.data = obj.attributes.get(db_attribute)
        self.obj = obj
        self.db_attribute = db_attribute
        self.cleanup()

    @property
    def all(self):
        """
        Returns a list of all keys in this object.
        """
        return list(self.data.keys())

    def ready(self, *args):
        """
        Checks whether all of the provided cooldowns are ready (expired). If a
        requested cooldown does not exist, it is considered ready.

        Args:
            *args (str): One or more cooldown names to check.
        Returns:
            bool: True if each cooldown has expired or does not exist.
        """
        return self.time_left(*args, use_int=True) <= 0

    def time_left(self, *args, use_int=False):
        """
        Returns the maximum amount of time left on one or more given cooldowns.
        If a requested cooldown does not exist, it is considered to have 0 time
        left.

        Args:
            *args (str): One or more cooldown names to check.
            use_int (bool): True to round the return value up to an int,
                False (default) to return a more precise float.
        Returns:
            float or int: Number of seconds until all provided cooldowns are
                ready. Returns 0 if all cooldowns are ready (or don't exist.)
        """
        now = time.time()
        cooldowns = [self.data[x] - now for x in args if x in self.data]
        if not cooldowns:
            return 0 if use_int else 0.0
        left = max(max(cooldowns), 0)
        return math.ceil(left) if use_int else left

    def add(self, cooldown, seconds):
        """
        Adds/sets a given cooldown to last for a specific amount of time.

        If this cooldown already exits, this call replaces it.

        Args:
            cooldown (str): The name of the cooldown.
            seconds (float or int): The number of seconds before this cooldown
                is ready again.
        """
        now = time.time()
        self.data[cooldown] = now + (max(seconds, 0) if seconds else 0)

    set = add

    def extend(self, cooldown, seconds):
        """
        Adds a specific amount of time to an existing cooldown.

        If this cooldown is already ready, this is equivalent to calling set. If
        the cooldown is not ready, it will be extended by the provided duration.

        Args:
            cooldown (str): The name of the cooldown.
            seconds (float or int): The number of seconds to extend this cooldown.
        Returns:
            float: The number of seconds until the cooldown will be ready again.
        """
        time_left = self.time_left(cooldown) + (seconds if seconds else 0)
        self.set(cooldown, time_left)
        return max(time_left, 0)

    def reset(self, cooldown):
        """
        Resets a given cooldown.

        Args:
            cooldown (str): The name of the cooldown.
        """
        if cooldown in self.data:
            del self.data[cooldown]

    def clear(self):
        """
        Resets all cooldowns.
        """
        self.data.clear()

    def cleanup(self):
        """
        Deletes all expired cooldowns. This helps keep attribute storage
        requirements small.
        """
        now = time.time()
        cooldowns = dict(self.data)
        keys = [x for x in cooldowns.keys() if cooldowns[x] - now < 0]
        if keys:
            for key in keys:
                del cooldowns[key]
            self.obj.attributes.add(self.db_attribute, cooldowns)
            self.data = self.obj.attributes.get(self.db_attribute)



"""Tests for text2html """

import unittest

import mock
from django.test import TestCase

from evennia.utils import ansi, text2html


class TestText2Html(TestCase):
    def test_format_styles(self):
        parser = text2html.HTML_PARSER
        self.assertEqual("foo", parser.format_styles("foo"))
        self.assertEqual(
            '<span class="color-001">red</span>foo',
            parser.format_styles(
                ansi.ANSI_UNHILITE + ansi.ANSI_RED + "red" + ansi.ANSI_NORMAL + "foo"
            ),
        )
        self.assertEqual(
            '<span class="bgcolor-001">red</span>foo',
            parser.format_styles(ansi.ANSI_BACK_RED + "red" + ansi.ANSI_NORMAL + "foo"),
        )
        self.assertEqual(
            '<span class="bgcolor-001 color-002">red</span>foo',
            parser.format_styles(
                ansi.ANSI_BACK_RED
                + ansi.ANSI_UNHILITE
                + ansi.ANSI_GREEN
                + "red"
                + ansi.ANSI_NORMAL
                + "foo"
            ),
        )
        self.assertEqual(
            'a <span class="underline">red</span>foo',
            parser.format_styles("a " + ansi.ANSI_UNDERLINE + "red" + ansi.ANSI_NORMAL + "foo"),
        )
        self.assertEqual(
            'a <span class="blink">red</span>foo',
            parser.format_styles("a " + ansi.ANSI_BLINK + "red" + ansi.ANSI_NORMAL + "foo"),
        )
        self.assertEqual(
            'a <span class="bgcolor-007 color-000">red</span>foo',
            parser.format_styles("a " + ansi.ANSI_INVERSE + "red" + ansi.ANSI_NORMAL + "foo"),
        )

    def test_remove_bells(self):
        parser = text2html.HTML_PARSER
        self.assertEqual("foo", parser.remove_bells("foo"))
        self.assertEqual(
            "a red" + ansi.ANSI_NORMAL + "foo",
            parser.remove_bells("a " + ansi.ANSI_BEEP + "red" + ansi.ANSI_NORMAL + "foo"),
        )

    def test_remove_backspaces(self):
        parser = text2html.HTML_PARSER
        self.assertEqual("foo", parser.remove_backspaces("foo"))
        self.assertEqual("redfoo", parser.remove_backspaces("a\010redfoo"))

    def test_convert_linebreaks(self):
        parser = text2html.HTML_PARSER
        self.assertEqual("foo", parser.convert_linebreaks("foo"))
        self.assertEqual("a<br> redfoo<br>", parser.convert_linebreaks("a\n redfoo\n"))

    def test_convert_urls(self):
        parser = text2html.HTML_PARSER
        self.assertEqual("foo", parser.convert_urls("foo"))
        self.assertEqual(
            'a <a href="http://redfoo" target="_blank">http://redfoo</a> runs',
            parser.convert_urls("a http://redfoo runs"),
        )

    def test_sub_mxp_links(self):
        parser = text2html.HTML_PARSER
        mocked_match = mock.Mock()
        mocked_match.groups.return_value = ["cmd", "text"]
        self.assertEqual(
            r"""<a id="mxplink" href="#" """
            """onclick="Evennia.msg(&quot;text&quot;,[&quot;cmd&quot;],{});"""
            """return false;">text</a>""",
            parser.sub_mxp_links(mocked_match),
        )

    def test_sub_text(self):
        parser = text2html.HTML_PARSER

        mocked_match = mock.Mock()

        mocked_match.groupdict.return_value = {"htmlchars": "foo"}
        self.assertEqual("foo", parser.sub_text(mocked_match))

        mocked_match.groupdict.return_value = {"htmlchars": "", "lineend": "foo"}
        self.assertEqual("<br>", parser.sub_text(mocked_match))

        parser.tabstop = 2
        mocked_match.groupdict.return_value = {
            "htmlchars": "",
            "lineend": "",
            "tab": "\t",
            "space": "",
        }
        self.assertEqual("  ", parser.sub_text(mocked_match))

        mocked_match.groupdict.return_value = {
            "htmlchars": "",
            "lineend": "",
            "tab": "\t\t",
            "space": " ",
            "spacestart": " ",
        }
        self.assertEqual("    ", parser.sub_text(mocked_match))

        mocked_match.groupdict.return_value = {
            "htmlchars": "",
            "lineend": "",
            "tab": "",
            "space": "",
            "spacestart": "",
        }
        self.assertEqual(None, parser.sub_text(mocked_match))

    def test_parse_tab_to_html(self):
        """Test entire parse mechanism"""
        parser = text2html.HTML_PARSER
        parser.tabstop = 4
        # single tab
        self.assertEqual(parser.parse("foo|>foo"), "foo    foo")

        # space and tab
        self.assertEqual(parser.parse("foo |>foo"), "foo     foo")

        # space, tab, space
        self.assertEqual(parser.parse("foo |> foo"), "foo      foo")

    def test_parse_html(self):
        self.assertEqual("foo", text2html.parse_html("foo"))
        self.maxDiff = None
        self.assertEqual(
            text2html.parse_html("|^|[CHello|n|u|rW|go|yr|bl|md|c!|[G!"),
            '<span class="blink bgcolor-006">'
            "Hello"
            '</span><span class="underline color-009">'
            "W"
            '</span><span class="underline color-010">'
            "o"
            '</span><span class="underline color-011">'
            "r"
            '</span><span class="underline color-012">'
            "l"
            '</span><span class="underline color-013">'
            "d"
            '</span><span class="underline color-014">'
            "!"
            '</span><span class="underline bgcolor-002 color-014">'
            "!"
            "</span>",
        )
